# coding: utf-8
###########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################

"""This module is a translation of octave ftseries developped by ESRF ID19 team
It contains the fasttomo global definitions, equivalent to
octave/defineGLOBALS.m
"""

__authors__ = ["C.Nemoz", "H.Payno"]
__license__ = "MIT"
__date__ = "19/03/2019"


import os
import subprocess
import glob
import datetime
from tomwer.core import settings
from tomwer.core import utils
from tomwer.core.utils.scanutils import MockEDF
from tomwer.core.process.task import Task
from tomwer.core.signal import Signal
from tomwer.core.utils import ftseriesutils
from tomwer.core.utils import logconfig
import logging
from tomwer.core.scan.scanbase import TomwerScanBase, _TomwerBaseDock
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.process.reconstruction.ftseries.params import ReconsParams
from silx.test.utils import temp_dir

logger = logging.getLogger(__name__)


class _Ftseries(Task, input_names=("data",), output_names=("data",)):
    """
    Ftseries is the class managing the reconstructions.
    Reconstructions are launched calling the octave fastomo3 scripts.

    Ftseries is able to run one reconstruction at the time only.
    Every requested reconstruction will be stored in a queue (`reconsStack`)
    and be launched one at the time on a dedicated thread.
    """

    # Note : when process is call we don't necessqrly expect to find an
    # 'octave_FT_params.h5' file at the folder level.
    # But updateReconsParam should always have a .h5 file defined

    sigScanReady = Signal(TomwerScanBase)
    """Signal emit when the reconstruction is finished"""

    def __init__(self, recons_stack=None, varinfo=None, inputs=None, name=None):
        Task.__init__(self, varinfo=varinfo, inputs=inputs, label=name)
        if "recons_params" in inputs:
            raise KeyError(
                "Do not use recons_params but `pyhst2_recons_params` instead"
            )
        recons_params = inputs.get("pyhst2_params", None)
        if not isinstance(recons_params, ReconsParams):
            raise TypeError(
                "pyhst2_recons_params should be provided as an instance"
                " of tomwer.core.process.reconstruction.ftseries.params.ReconsParams and not {}".format(
                    type(recons_params)
                )
            )
        self._recons_params = recons_params

        """Reconstruction parameters for tomo reconstruction"""
        self.dry_run = inputs.get("dry_run", False)
        """execute the pyhst reconstruction or only create the .par file"""

        self._scan = inputs.get("data", None)
        self._exploreForH5File = inputs.get("explore_for_h5_file", False)
        """If True then look in the scan folder and if contain a .h5 try to
        load to update the reconstruction parameters from it"""
        self._mockMode = inputs.get("mock_mode", False)
        """If we want to skip pyhst reconstruction and create some random
        slice reconstruction"""
        self.reconsStack = recons_stack
        if self.reconsStack is not None:
            self.reconsStack.sigReconsMissParams.connect(self.updateReconsParam)
            self.reconsStack.sigReconsFinished.connect(self._signalReconsReady)

    @property
    def recons_params(self):
        return self._recons_params

    def set_configuration(self, properties):
        if "_rpSetting" in properties:
            self.recons_params.setStructs(properties["_rpSetting"])

    def set_dry_run(self, dry_run):
        self.dry_run = dry_run

    def set_recons_params(self, recons_param):
        assert isinstance(recons_param, ReconsParams)
        self._recons_params = recons_param
        # TODO: this should be update to use am update instead

    def setH5Exploration(self, b):
        self._exploreForH5File = b

    def updatePath(self, path):
        """
        update self._scan and try to load the configuration file for octave

        :param path:the path to the new folder to observe
        """
        if self._exploreForH5File:
            h5file = ftseriesutils.tryToFindH5File(path, "newest")
            "h5 file is the file to load to update the reconstruction parameters"
            if h5file is not None:
                try:
                    self.load_recons_params(h5file)
                except Exception:
                    logger.warning(
                        "Fail to load reconstruction parameters " "from %s" % h5file
                    )
                else:
                    logger.info("Reconstruction parameters loaded from " "%s" % h5file)

        self._scan = ScanFactory.create_scan_object(scan_path=path)

    def pathReceived(self, pathToTheScanCompleted):
        """Callback function when the path of the file to scan is modify
        The default behavior here is to run a first reconstruction without
        waiting any user modifications

        :param pathToTheScanCompleted:the path to the new folder to observe"""
        if pathToTheScanCompleted is None:
            return

        logger.info("%s received" % pathToTheScanCompleted)

        if type(pathToTheScanCompleted) is str:
            _scan = ScanFactory.create_scan_object(scan_path=pathToTheScanCompleted)
        else:
            _scan = pathToTheScanCompleted

        # behavior : when receiving a new scan, reconstruct it right away
        return self.process(scan=_scan)

    def updateCurrentFTSeries(self, scan):
        """Change the Ftseries we want to reconstruct"""
        if scan is not None:
            info = "ask for FTSerie scan %s" % scan.path
            logger.info(
                info,
                extra={
                    logconfig.DOC_TITLE: self._scheme_title,
                    logconfig.SCAN_ID: scan.path,
                },
            )

            self._scan = scan

    def run(self):
        """Call the core function 'run_reconstruction' of the ftseries script
        which will call octave to process reconstruction
        """
        scan = self.inputs.data
        if scan is None:
            return

        if type(scan) is str:
            # if necessary updatePath
            if self._scan is None or scan != self._scan.path:
                self.updatePath(scan)
        elif isinstance(scan, TomwerScanBase):
            self._scan = scan
        elif type(scan) is dict:
            self._scan = ScanFactory.create_scan_object_frm_dict(scan)
        else:
            raise TypeError(
                "scan should be a path to a valid scan or an " "instance of TomoScan"
            )

        if (
            self._scan is not None
            and self._scan.path is not None
            and os.path.isdir(self._scan.path)
        ):

            def copy_recons_params():
                """as we have a widget able to define some dark ref parameter
                we have first to copy this parameter value if any"""
                if self._scan.ftseries_recons_params is None:
                    self._scan.ftseries_recons_params = ReconsParams(empty=True)

                self._recons_params.copy(self._scan.ftseries_recons_params.dkrf)

            # first copy some recons params from the input edf tomo scan
            copy_recons_params()

            # then update the scan ftseries_recons_params
            self._scan.ftseries_recons_params.copy(self.recons_params)
            assert (
                self._scan.ftseries_recons_params.axis.do_axis_correction
                == self.recons_params.axis.do_axis_correction
            )

            if self._mockMode:
                # mock are not made within a thread
                logger.info("mocking reconstruction")
                MockEDF.mockReconstruction(self._scan.path)
            elif self.reconsStack is None:
                # run reconstruction without stack
                scan.ftseries_recons_params.copy(self.recons_params)
                inputs = {
                    "pyhst2_params": scan.ftseries_recons_params.to_dict(),
                    "dry_run": self.dry_run,
                }

                run_reconstruction_from_recons_params(scan=self._scan, inputs=inputs)
                self._signalReconsReady(scan=self._scan)
            else:
                # note: _signalReconsReady will be called by the slot associated
                # to the stack signals
                self.reconsStack.add(
                    scan=self._scan,
                    inputs=self.recons_params.to_unique_recons_set(as_to_dict=True),
                    schemeTitle=self._scheme_title,
                    dry_run=self.dry_run,
                )
            self.register_output(key="data", value=self._scan)
            if self._return_dict:
                self.outputs.data = self._scan.to_dict()
            else:
                self.outputs.data = self._scan

    def _infoMissingStrucVar(missingStructures, missingVariables):
        m = "File is missing some structures and/or variables."
        m += "Values of those has been setted to default values."
        m += "Please make sure they are correct."
        m += "\nMissing structures: " + missingStructures
        m += "\nMissing variables: " + missingVariables
        print(m)

    def _signalReconsReady(self, scan):
        assert isinstance(scan, TomwerScanBase)

        info = "scan %s reconstructed" % scan.path
        logger.processEnded(
            info,
            extra={
                logconfig.DOC_TITLE: self._scheme_title,
                logconfig.SCAN_ID: scan.path,
            },
        )

        # if some volraw or volfloat are present at the same level, let
        # synchronize them.
        if settings.isOnLbsram(scan) is True:
            volfloat = os.path.join(os.path.dirname(scan.path), "volraw")
            volraw = os.path.join(os.path.dirname(scan.path), "volfloat")

            for _folder in (volraw, volfloat):

                if os.path.exists(_folder) and os.path.isdir(_folder):
                    target = _folder.replace(
                        settings.get_lbsram_path(), settings.get_dest_path(), 1
                    )
                    logger.info(
                        "start synchronization between %s and %s" % (_folder, target)
                    )
                    if not os.path.exists(target):
                        try:
                            os.makedirs(target)
                        except Exception as e:
                            logger.error(e)
                        else:
                            os.chmod(target, 0o774)
        self.sigScanReady.emit(scan)

    def askUserH5File(self):
        filePath = None
        while filePath is None:
            out = input("please give the path to the h5 file : \n")
            if not os.path.isfile(filePath):
                warning = "given path " + out
                warning += " is not a directory, please give a valid directory"
                logger.warning(warning)
                out = None
        return out

    def askUserAndLoad(self):
        """Process launch by activing the Load button"""
        f = self.askUserH5File()
        if f is not None:
            self.load_recons_params(f)

    def load_recons_params(self, h5file):
        assert os.path.isfile(h5file)
        from .params.fastsetupdefineglobals import FastSetupAll  # avoid cyclic import

        fsdg = FastSetupAll()
        fsdg.readAll(h5file, 3.8)
        self.recons_params.copy(fsdg.structures)

    @staticmethod
    def save_recons_params(recons_params: dict, h5File, displayInfo=True):
        if not isinstance(recons_params, dict):
            raise ValueError("recons_params is expected to be a dict")
        ftseriesutils.saveH5File(
            structs=recons_params, h5File=h5File, displayInfo=displayInfo
        )

    def save(self, h5File, displayInfo=True):
        """Function to overwrite the reconstruction parameters into the h5 file"""
        self.save_recons_params(
            recons_params=self._recons_params.to_dict(),
            h5File=h5File,
            displayInfo=displayInfo,
        )

    def setMockMode(self, b):
        """If the mock mode is activated then during reconstruction won't call
        Octave script for reconstruction but will generate some output files
        according to convention

        :param boolean b: True if we want to active the mock mode
        """
        self._mockMode = b

    def updateReconsParam(self, ftserie):
        if ftserie is None:
            return

        scan = ftserie
        if type(ftserie) is str:
            scan = ScanFactory.create_scan_object(ftserie)
        elif type(ftserie) is dict:
            scan = ScanFactory.create_scan_object_frm_dict(ftserie)
        elif isinstance(scan, _TomwerBaseDock):
            scan = ftserie.instance
        elif isinstance(scan, TomwerScanBase):
            scan = ftserie
        else:
            raise ValueError("Type not managed")
        assert scan is not None
        self.updateCurrentFTSeries(scan)

        # if on lbsram and in low memory then skip it
        if (
            settings.isOnLbsram(scan)
            and utils.isLowOnMemory(settings.get_lbsram_path()) is True
        ):
            # if computer is running into low memory in lbsram skip reconstruction
            mess = "low memory, skip reconstruction for " + scan.path
            logger.processSkipped(mess)
            self._signalReconsReady(self._scan.scanID)
        else:
            return self._askForReconsParamValidation()

    def _askForReconsParamValidation(self):
        logger.warning(
            "Reconstruction parameters validation is only develop "
            "with gui interaction for now. skip it."
        )
        return self.process(self._scan.path)

    def setForceSync(self, b):
        """
        Force synchronisation of the reconstruction

        :param b: True if we want to block ftseries during reconstruction
        """
        if self.reconsStack is not None:
            self.reconsStack.setForceSync(b)


class H5NoFileException(Exception):
    """Exception launch when no .h5 is found by ftseries"""

    pass


def _subprocess_run(*popenargs, _input=None, timeout=None, check=False, **kwargs):
    if _input is not None:
        if "stdin" in kwargs:
            raise ValueError("stdin and input arguments may not both be used.")
        kwargs["stdin"] = subprocess.PIPE

    process = subprocess.Popen(*popenargs, **kwargs)
    try:
        stdout, stderr = process.communicate(_input)
    except Exception:
        process.kill()
        if timeout is not None:
            process.wait(timeout)
        else:
            process.wait()
        raise
    retcode = process.poll()
    if check and retcode:
        raise subprocess.CalledProcessError(
            retcode, process.args, output=stdout, stderr=stderr
        )
    return retcode, stdout, stderr


def run_reconstruction_from_recons_params(scan: TomwerScanBase, inputs: dict) -> None:
    """

    ..note: in this case the recons_param should be unique (not a set/List
            of ReconsParams)

    :param TomwerScanBase scan:
    :param dict inputs:
    """
    if not isinstance(scan, TomwerScanBase):
        raise TypeError(f"{scan} is expected to be an instance of {TomwerScanBase}")
    if not isinstance(inputs, dict):
        raise TypeError(f"{inputs} is expected to be an instance of {dict}")

    from tomwer.core.process.reconstruction.pyhst.pyhstcaller import PyHSTCaller

    if "data" in inputs:
        logger.warning(
            f"'data key' will be overwrite (replace {inputs['data']} by {scan})"
        )
    inputs["data"] = scan

    pyhstcaller = PyHSTCaller(inputs=inputs)
    pyhstcaller.run()


def run_reconstruction(directory, h5file):
    """Launch a reconstruction

    :param str directory: the acquisition directory
    :param str h5file: the h5 file containing the reconstruction parameters
    """
    with temp_dir() as tmp:
        # initialise global structures with default values
        assert os.path.isdir(directory)

        # manage all sub directories
        dirs = []

        aux = glob.glob(directory)
        laux = len(aux)
        for i in range(laux):
            dirs.append(aux[i])

        # now : new behavior : each directory for which we are running a
        # reconstruction should have an .h5 file

        if h5file is None or not os.path.isfile(h5file):
            raise H5NoFileException("No h5 file found for reconstruction")

        # version march python2/3 avec octave 3.6
        # fail avec octave 3.8

        octavexe = tmp + "/octseries"
        try:
            with open(octavexe, "wb") as file_:
                file_.write(bytes("#!/usr/bin/octave\n", "UTF-8"))
                file_.write(bytes("\n", "UTF-8"))
                file_.write(bytes("load_pars " + h5file + "\n", "UTF-8"))
                file_.write(bytes("global GET_MY_FT\n", "UTF-8"))
                file_.write(bytes("global GENERATOR\n", "UTF-8"))
                file_.write(bytes("GET_MY_FT = 0;\n", "UTF-8"))
                file_.write(bytes("global GENERATOR\n", "UTF-8"))
                file_.write(bytes("GENERATOR = 'tomwer';\n", "UTF-8"))
                file_.write(bytes("cd " + dirs[0] + "\n", "UTF-8"))
                file_.write(bytes("fasttomo3\n", "UTF-8"))
                file_.write(bytes("\n", "UTF-8"))
                file_.close()
            logger.info("creation of %s succeded" % octavexe)
        except Exception:
            logger.error("fail writing of %s" % octavexe)
            raise

        timeout = 15 * 60  # in sec
        try:
            # TODO : set write only to the current user security issue.
            os.chmod(octavexe, 0o775)
            octave_log = directory + "/octave.log"
            with open(octave_log, "ab") as _file:
                mydate = datetime.datetime.now()
                _file.write(
                    bytes(
                        "\n\n\n===== OCTAVE process started on " + str(mydate), "UTF-8"
                    )
                )

            with open(octave_log, "ab") as _file:
                retcode, stdout, stderr = _subprocess_run(
                    octavexe,
                    timeout=timeout,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )
                if stdout is not None:
                    print(stdout.decode("utf-8"))
                    _file.write(stdout)
                if retcode > 0 and stderr is not None:
                    _file.write(stderr)
                    logger.error("error append during execution of %s" % octavexe)
                    logger.error(stderr.decode("utf-8"))
                else:
                    logger.info("succeeded execution of %s" % octavexe)
        except subprocess.TimeoutExpired:
            logger.warning(
                "reconstruction for %s take too long (%s)" "" % (octavexe, timeout)
            )
        except Exception:
            logger.error("fail execution of %s" % octavexe)
            raise

        return h5file


def getInfoFile(dir):
    files = glob.glob(dir + "*.info")
    if dir.startswith(settings.get_lbsram_path()):
        seconddir = dir.rstrip(settings.get_lbsram_path())
        if os.path.isdir(seconddir):
            files += glob.glob(seconddir + "*.info")
    return files


# TODO: include this in _Ftseries
class Ftseries(_Ftseries):
    def __init__(self, varinfo=None, inputs=None, name=None):
        recons_params = inputs.get("pyhst2_params", None)
        # lazy loading
        from tomwer.core.process.reconstruction.ftseries.params import ReconsParams

        recons_params = recons_params or ReconsParams()
        inputs["pyhst2_params"] = recons_params
        inputs["force_sync"] = True

        _Ftseries.__init__(
            self, varinfo=varinfo, inputs=inputs, recons_stack=None, name=name
        )



# /*##########################################################################
# Copyright (C) 2016-2017 European Synchrotron Radiation Facility
#
# This file is part of the PyMca X-ray Fluorescence Toolkit developed at
# the ESRF by the Software group.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__author__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "24/01/2017"

from silx.gui import qt
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui import stackplot
from tomwer.io.utils import get_default_directory
import os

import logging

logger = logging.getLogger(__name__)


class ScanWidget(qt.QWidget):
    """
    Widget to display all scan information

    :param parent: the qt parent of the widget
    :param canLoadOtherScan: can we load an other scan
    """

    def __init__(self, parent=None, canLoadOtherScan=False):

        qt.QWidget.__init__(self, parent)
        self.ftseriereconstruction = None

        self.canLoadOtherScan = canLoadOtherScan
        self.__loadGUI()

    def __loadGUI(self):
        """Function loading the GUI. Not done on the constructor to avoid memory
        charge
        """

        layout = qt.QVBoxLayout()
        # add the scan ID label or folder getter
        self._folderSelection = self.__createFolderSelection()
        self._folderSelection.setSizePolicy(
            qt.QSizePolicy.Minimum, qt.QSizePolicy.Minimum
        )
        layout.addWidget(self._folderSelection)
        # create viewers
        self.stackImageViewerTab = qt.QTabWidget(self)
        self.stackImageViewerTab.setSizePolicy(
            qt.QSizePolicy.Expanding, qt.QSizePolicy.Expanding
        )
        layout.addWidget(self.stackImageViewerTab)
        self.setLayout(layout)

        self.stackImageViewerTab.setLayout(qt.QVBoxLayout())
        self.stackImageViewerTab.layout().setContentsMargins(0, 0, 0, 0)
        self._stackImageViewerScan = stackplot._QImageFileStackPlot(self)
        # self._stackImageViewerScan.setTranspose(True)
        self.stackImageViewerTab.addTab(self._stackImageViewerScan, "Reconstruction")

        self._stackImageViewerRadio = stackplot._QImageFileStackPlot(self)
        self.stackImageViewerTab.addTab(self._stackImageViewerRadio, "Radios")

        self.loaded = True

    def __createFolderSelection(self):
        layoutScanId = qt.QHBoxLayout()
        widget = qt.QWidget(self)
        widget.setLayout(layoutScanId)
        self.folderName = qt.QLineEdit(widget)
        layoutScanId.addWidget(self.folderName)
        self.button = qt.QPushButton("Select folder", parent=widget)
        self._rootFolderSelection = None
        layoutScanId.addWidget(self.button)
        self.button.clicked.connect(self._getFolder)
        # self.folderName.textChanged.connect(self._loadNewSerie)

        if not self.canLoadOtherScan:
            self.button.hide()
            self.folderName.setEnabled(False)

        return widget

    def setRootFolderSelection(self, path):
        """Set the path open when selection a folder to be diaplyed

        :param str path: root path to be open when asking for a new folder
        """
        self._rootFolderSelection = path

    def getCurrentScanFolder(self):
        """Return the folder we want to scan"""
        return self.folderName.text()

    def _loadNewSerie(self, scan):
        """
        Load a new serie from the given scan

        :param TomwerScanBase scan: scan to be displayed
        """
        if scan is None:
            return
        assert isinstance(scan, TomwerScanBase)
        self.updateData(scan)

    def updateData(self, scan):
        """
        Update the current ftSerieReconstruction displayed

        :param FtserieReconstruction scan: the new serie to be
        displayed
        """
        if scan is None or scan.path is None:
            return

        assert type(scan.path) is str

        self.ftseriereconstruction = scan
        self.ftseriereconstruction.update()
        self.loaded = False
        self.folderName.setText(str(scan))
        self._stackImageViewerRadio.setImages(self.ftseriereconstruction.projections)
        self._stackImageViewerScan.setImages(
            self.ftseriereconstruction.get_reconstructions_urls()
        )

    def _getFolder(self):
        """
        Call back when the user want to change the folder to validate
        """

        defaultDirectory = self._rootFolderSelection or get_default_directory()
        currentSettedFolder = self.getCurrentScanFolder()

        if currentSettedFolder is not None and os.path.isdir(currentSettedFolder):
            defaultDirectory = currentSettedFolder

        dialog = qt.QFileDialog(self, directory=defaultDirectory)
        dialog.setFileMode(qt.QFileDialog.DirectoryOnly)

        if not dialog.exec_():
            dialog.close()
            return None

        self.folderName.setText(dialog.selectedFiles()[0])

    def clear(self):
        self.ftseriereconstruction = None

        self.loaded = False
        self.folderName.setText("")
        self._stackImageViewerRadio.clear()
        self._stackImageViewerScan.clear()

    def showActiveImage(self):
        self._stackImageViewerRadio.updateActiveImage()
        self._stackImageViewerScan.updateActiveImage()


class ScanWidgetValidation(ScanWidget):
    """
    This is the same as ScanWidget but include a widget button which will
    emit signals such as validated, canceled...

    :param parent: the qt parent of the widget
    :param canLoadOtherScan: can we load an other scan
    """

    def __init__(self, parent=None, canLoadOtherScan=False):
        ScanWidget.__init__(self, parent=parent, canLoadOtherScan=canLoadOtherScan)

        self.validationWidget = ValidationWidget(
            parent=self, ftseriereconstruction=self.ftseriereconstruction
        )
        self.validationWidget.setVisible(not canLoadOtherScan)
        self.validationWidget.setSizePolicy(
            qt.QSizePolicy.Minimum, qt.QSizePolicy.Minimum
        )

        self.layout().addWidget(self.validationWidget)

    def updateData(self, scan):
        ScanWidget.updateData(self, scan)
        if scan is not None:
            self.validationWidget.ftseriereconstruction = scan


class ValidationWidget(qt.QGroupBox):
    """
    Class containing all the validation buttons
    and sending signals when they are pushed

    :param QObject parent: the parent of the QTabWidget
    :param :class:`.FtserieReconstruction`: the scan to display
    """

    sigValidateScan = qt.Signal(str)
    sigCancelScan = qt.Signal(str)
    sigRedoAcquisitionScan = qt.Signal(TomwerScanBase)
    sigChangeReconstructionParametersScan = qt.Signal(TomwerScanBase)

    def __init__(self, ftseriereconstruction, parent=None):
        qt.QGroupBox.__init__(self, title="Validate manually", parent=parent)

        self.setCheckable(True)
        self.ftseriereconstruction = ftseriereconstruction
        layout = qt.QGridLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

        #  validate button
        self.validateButton = qt.QPushButton("Validate")
        style = qt.QApplication.style()
        self.validateButton.setIcon(style.standardIcon(qt.QStyle.SP_DialogApplyButton))
        self.validateButton.pressed.connect(self.__validated)
        layout.addWidget(self.validateButton, 0, 2)

        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        layout.addWidget(spacer, 0, 1)

        #  cancel button
        self.cancelButton = qt.QPushButton("Cancel")
        self.cancelButton.setIcon(style.standardIcon(qt.QStyle.SP_DialogCancelButton))
        self.cancelButton.pressed.connect(self.__canceled)
        layout.addWidget(self.cancelButton, 0, 0)
        # for now this button is hidden because the behavior haven't been
        # clearly defined
        # self.cancelButton.hide()

        #  change reconstruction parameters button
        self.changeReconsParamButton = qt.QPushButton(
            "Change reconstruction parameters"
        )
        self.changeReconsParamButton.setIcon(
            style.standardIcon(qt.QStyle.SP_FileDialogContentsView)
        )
        self.changeReconsParamButton.pressed.connect(
            self.__updateReconstructionParameters
        )
        layout.addWidget(self.changeReconsParamButton, 2, 0)

    def setEnabled(self, b):
        self.validateButton.setEnabled(b)
        self.cancelButton.setEnabled(b)
        self.changeReconsParamButton.setEnabled(b)

    def __validated(self):
        """Callback when the validate button is pushed"""
        self.sigValidateScan.emit("")

    def __canceled(self):
        """Callback when the cancel button is pushed"""
        if self.ftseriereconstruction is not None:
            self.setEnabled(False)
            self.sigCancelScan.emit(self.ftseriereconstruction.path)

    def __redoacquisition(self):
        """Callback when the redo acquisition button is pushed"""
        if self.ftseriereconstruction is not None:
            self.setEnabled(False)
            self.sigRedoAcquisitionScan.emit(self.ftseriereconstruction)

    def __updateReconstructionParameters(self):
        """Callback when the change reconstruction button is pushed"""
        if self.ftseriereconstruction is not None:
            self.setEnabled(False)
            self.sigChangeReconstructionParametersScan.emit(self.ftseriereconstruction)


class ImageStackViewerValidator(ScanWidgetValidation):
    """
    Widget to visualize a stack of image

    :param QObject parent: the parent of the QTabWidget
    :param :class:`.FtserieReconstruction`: the scan to display
    """

    _sizeHint = qt.QSize(600, 600)

    def __init__(self, parent=None, ftseries=None):
        ScanWidgetValidation.__init__(self, parent)

        self._scanWidgetLayout = qt.QVBoxLayout()
        self.layout().setContentsMargins(0, 0, 0, 0)

        if ftseries is not None:
            self.addScan(ftseries)

    def addScan(self, ftseriereconstruction, entry=None):
        """function called for showing infomration about a new reconstruction

        :param FtserieReconstruction ftseriereconstruction:
            contains all information about the reconstruciton (scan path,
            reconstruction path ... )
        """
        if not type(ftseriereconstruction) is TomwerScanBase:
            raise RuntimeError(
                "Update error can't manage a type different than\
                                 FtserieReconstruction"
            )

        assert ftseriereconstruction is not None
        assert ftseriereconstruction.path is not None

        logger.info(
            "Scan validator received %s to be validated" % ftseriereconstruction.path
        )

        self.lastReconstructionReceived = ftseriereconstruction
        self._updateData(self.lastReconstructionReceived)

    def updateFromPath(self, path, entry=None):
        """Show the reconstruction from a path"""
        if not os.path.isdir(path):
            raise RuntimeError("givem path %s is not a directory" % path)

        self.addScan(ScanFactory.create_scan_object(path, entry=entry))

    def __updateScanTabsID(self):
        if hasattr(self, "lastReconstructionReceived") and (
            self.lastReconstructionReceived is not None
        ):
            # update FtSerie in scan to treat tab
            self.updateData(self.lastReconstructionReceived)
            # update folder path in other tab
            assert self.OTHER_TAB in self.tabsWidget
            if self.lastReconstructionReceived.path is not None:
                self.setRootFolderSelection(
                    os.path.dirname(self.lastReconstructionReceived.path)
                )

    def sizeHint(self):
        """Return a reasonable default size for usage in :class:`PlotWindow`"""
        return self._sizeHint


def setUp():
    import fabio.edfimage
    import tempfile
    import numpy

    folder = tempfile.mkdtemp()
    basename = os.path.basename(folder)

    # create projections
    for i in range(20):
        f = tempfile.mkstemp(
            prefix=basename, suffix=str("_" + str(i) + ".edf"), dir=folder
        )
        data = numpy.array(numpy.random.random(200 * 200))
        data.shape = (200, 200)
        edf_writer = fabio.edfimage.EdfImage(data=data, header={"tata": "toto"})
        edf_writer.write(f[1])

    # create reconstruction
    for i in range(5):
        f = tempfile.mkstemp(
            prefix=basename, suffix=str("_slice_" + str(i) + ".edf"), dir=folder
        )
        data = numpy.zeros((200, 200))
        data[:: i + 2, :: i + 2] = 1.0
        edf_writer = fabio.edfimage.EdfImage(data=data, header={"tata": "toto"})
        edf_writer.write(f[1])

    f = tempfile.mkstemp(
        prefix=basename, suffix=str("_slice_" + str(6) + ".edf"), dir=folder
    )
    data = numpy.zeros((200, 200))
    data[50:150, 50:150] = 1.0
    edf_writer = fabio.edfimage.EdfImage(data=data, header={"tata": "toto"})
    edf_writer.write(f[1])

    return folder



# coding: utf-8
# ##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################
"""
Simple logs levels definition relative to process status
"""


__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "02/10/2020"

import logging


PROCESS_STARTED_NAME = "PROCESS_STARTED"
PROCESS_STARTED_LEVEL = 32

PROCESS_SUCCEED_NAME = "PROCESS_SUCCEED"
PROCESS_SUCCEED_LEVEL = 33

PROCESS_FAILED_NAME = "PROCESS_FAILED"
PROCESS_FAILED_LEVEL = 38

PROCESS_ENDED_NAME = "PROCESS_ENDED"
"""String name of the process ended"""

PROCESS_ENDED_LEVEL = 35
"""Level of the process ended"""

PROCESS_SKIPPED_NAME = "PROCESS_SKIPPED"
"""String name of the process skipped"""

PROCESS_SKIPPED_LEVEL = 36
"""Level of the process skipped"""

PROCESS_INFORM_NAME = "PROCESS_INFORM"
"""Name of the process information"""

PROCESS_INFORM_LEVEL = 31
"""Level for information send to graylog"""

logging.addLevelName(PROCESS_STARTED_LEVEL, PROCESS_STARTED_NAME)
logging.addLevelName(PROCESS_SUCCEED_LEVEL, PROCESS_SUCCEED_NAME)
logging.addLevelName(PROCESS_FAILED_LEVEL, PROCESS_FAILED_NAME)
logging.addLevelName(PROCESS_ENDED_LEVEL, PROCESS_ENDED_NAME)
logging.addLevelName(PROCESS_SKIPPED_LEVEL, PROCESS_SKIPPED_NAME)
logging.addLevelName(PROCESS_INFORM_LEVEL, PROCESS_INFORM_NAME)


def processEnded(self, message, *args, **kws):
    if self.isEnabledFor(PROCESS_ENDED_LEVEL):
        self._log(PROCESS_ENDED_LEVEL, message, args, **kws)


def processSkipped(self, message, *args, **kws):
    if self.isEnabledFor(PROCESS_SKIPPED_LEVEL):
        self._log(PROCESS_SKIPPED_LEVEL, message, args, **kws)


def inform(self, message, *args, **kws):
    if self.isEnabledFor(PROCESS_INFORM_LEVEL):
        self._log(PROCESS_INFORM_LEVEL, message, args, **kws)


def processSucceed(self, message, *args, **kws):
    if self.isEnabledFor(PROCESS_SUCCEED_LEVEL):
        self._log(PROCESS_SUCCEED_LEVEL, message, args, **kws)


def processFailed(self, message, *args, **kws):
    if self.isEnabledFor(PROCESS_FAILED_LEVEL):
        self._log(PROCESS_FAILED_LEVEL, message, args, **kws)


def processStarted(self, message, *args, **kws):
    if self.isEnabledFor(PROCESS_STARTED_LEVEL):
        self._log(PROCESS_STARTED_LEVEL, message, args, **kws)


logging.Logger.processEnded = processEnded
logging.Logger.processSkipped = processSkipped
logging.Logger.inform = inform
logging.Logger.processSucceed = processSucceed
logging.Logger.processFailed = processFailed
logging.Logger.processStarted = processStarted



# /*##########################################################################
# Copyright (C) 2017 European Synchrotron Radiation Facility
#
# This file is part of the PyMca X-ray Fluorescence Toolkit developed at
# the ESRF by the Software group.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/
"""Some widget construction to check if a sample moved"""

__author__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "01/06/2018"


from silx.gui import qt
from silx.io import dictdump
import logging
from tomwer.core.process.reconstruction.lamino.tofu import (
    SCAN_TYPES,
    SCAN_TYPES_I,
    FFCWhen,
)
from tomwer.core.process.reconstruction.lamino.tofu import _retrieve_opts_recons_cmd
from tomwer.core.utils.char import PSI_CHAR
from tomwer.gui import icons
from tomwer.gui.reconstruction.lamino.tofu import settings
from .TofuOptionLoader import _TofuOptionLoader, _getterSetter
from .projections import InputProjectionsScrollArea
from .tofuexpert import ExpertTofuWidget
from .tofuoutput import OutputTofuWidget

_logger = logging.getLogger(__name__)


class TofuWindow(qt.QMainWindow):
    """
    Widget used to interface lamino reconstruction using tofu
    """

    def __init__(self, parent):
        qt.QMainWindow.__init__(self, parent)
        self.setWindowFlags(qt.Qt.Widget)
        # main widget
        self._mainWidget = TofuWidget(parent=self)
        self.setCentralWidget(self._mainWidget)

        # expose API
        self.save = self._mainWidget.save
        self.load = self._mainWidget.load
        self.loadFromScan = self._mainWidget.loadFromScan
        self.getParameters = self._mainWidget.getParameters
        self.setParameters = self._mainWidget.setParameters
        self.getAdditionalRecoOptions = self._mainWidget.getAdditionalRecoOptions
        self.setAdditionalRecoOptions = self._mainWidget.setAdditionalRecoOptions
        self.getAdditionalPreprocessOptions = (
            self._mainWidget.getAdditionalPreprocessOptions
        )
        self.setAdditionalPreprocessOptions = (
            self._mainWidget.setAdditionalPreprocessOptions
        )
        self.removeOutputDir = self._mainWidget.removeOutputDir
        self.setRemoveOutputDir = self._mainWidget.setRemoveOutputDir

        # menu
        style = qt.QApplication.style()
        saveIcon = style.standardIcon(qt.QStyle.SP_DialogSaveButton)
        self._saveAction = qt.QAction(saveIcon, "save", self)
        self._saveAction.setToolTip("save configuration")
        self._saveAction.setShortcut(qt.QKeySequence.Save)
        self.menuBar().addAction(self._saveAction)
        self._saveAction.triggered.connect(self.save)

        loadIcon = style.standardIcon(qt.QStyle.SP_DialogOpenButton)
        self._loadAction = qt.QAction(loadIcon, "load", self)
        self._loadAction.setToolTip("load configuration")
        self._loadAction.setShortcut(qt.QKeySequence.Open)
        self.menuBar().addAction(self._loadAction)
        self._loadAction.triggered.connect(self.load)


class TofuWidget(_TofuOptionLoader, qt.QWidget):
    _OPT_RECO_SECTION_ID = "options_reco"
    """Variable used to define the section name when saving the configuration
    into dict dump"""
    _OPT_PREPROCESS_SECTION_ID = "options_preprocess"

    _TOMWER_CONFIG = "tomwer_configuration"
    "tomwer processing configuration"

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())
        self._scanTypeWidget = ScanTypeWidget(parent=self)
        self.layout().addWidget(self._scanTypeWidget)

        self._tabs = _TofuTab(parent=self)
        self.layout().addWidget(self._tabs)

        _TofuOptionLoader.__init__(
            self,
            options={
                "z-parameter": _getterSetter(
                    getter=self._scanTypeWidget.getScanType,
                    setter=self._scanTypeWidget.setScanType,
                )
            },
            childs=[self._tabs],
        )

        # expose API
        self.getAdditionalRecoOptions = self._tabs.getAdditionalRecoOptions
        self.getAdditionalPreprocessOptions = self._tabs.getAdditionalPreprocessOptions
        self.removeOutputDir = self._tabs._outputWidget._outputWidget.removeIfExist
        self.setRemoveOutputDir = (
            self._tabs._outputWidget._outputWidget.setRemoveIfExist
        )

        # connect signals
        self._scanTypeWidget._scanType.currentIndexChanged.connect(
            self._scanTypeChanged
        )
        self._tabs._outputWidget._volumeAngleGrp._grpPlane.sigPlaneChanged.connect(
            self._scanTypeChanged
        )
        # update display according to the current scan type
        self._scanTypeChanged()

    def _scanTypeChanged(self, *args, **kwargs):
        scan_type = self._scanTypeWidget._scanType.currentText()
        x_center = 0
        # change step size type
        if scan_type == "slice stack":
            _type = "(pixel)"
            # z_center = self._tabs._inputWidget._centeringWidget.getZCenter()
            # _range = (z_center - settings.SLICE_STACK_RANGE_HS,
            #           z_center + settings.SLICE_STACK_RANGE_HS)
            _range = -settings.SLICE_STACK_RANGE_HS, settings.SLICE_STACK_RANGE_HS
            _step_size = settings.SLICE_STACK_STEP_SIZE
            _nb_slices = settings.SLICE_STACK_NB_SLICES
        elif scan_type == "rotation center":
            _type = "(pixel)"
            x_center = self._tabs._inputWidget.centeringWidget.getXCenter()
            _range = (
                x_center - settings.ROT_CENTER_RANGE_HS,
                x_center + settings.ROT_CENTER_RANGE_HS,
            )
            _step_size = settings.ROT_CENTER_STEP_SIZE
            _nb_slices = settings.ROT_CENTER_NB_SLICES
        elif scan_type == "lamino angle":
            _type = "(degree)"
            lamino_angle = self._tabs._inputWidget.rotationAngle.getLaminoAngle()
            x_center = lamino_angle
            _range = (
                lamino_angle - settings.LAMINO_ANG_RANGE_HS,
                lamino_angle + settings.LAMINO_ANG_RANGE_HS,
            )
            _step_size = settings.LAMINO_ANG_STEP_SIZE
            _nb_slices = settings.LAMINO_ANG_NB_SLICES
        elif scan_type == PSI_CHAR + " angle":
            _type = "(degree)"
            psi_angle = self._tabs._inputWidget.rotationAngle.getPsiAngle()
            x_center = psi_angle
            # _range = (psi_angle - settings.PSI_ANG_RANGE_HS,
            #           psi_angle + settings.PSI_ANG_RANGE_HS)
            _range = -settings.PSI_ANG_RANGE_HS, settings.PSI_ANG_RANGE_HS
            _step_size = settings.PSI_ANG_STEP_SIZE
            _nb_slices = settings.PSI_ANG_NB_SLICES
        else:
            raise NotImplementedError("")

        _stepSizeAndRangeWidget = self._tabs._outputWidget._stepSizeAndRange
        _stepSizeAndRangeWidget._setStepSizeType(_type)
        _stepSizeAndRangeWidget._z_label.setVisible(scan_type == "slice stack")
        _stepSizeAndRangeWidget._zSB.setVisible(scan_type == "slice stack")
        # change step range and size
        _stepSizeAndRangeWidget.setRegion(region=(_range[0], _range[1], _step_size))
        _stepSizeAndRangeWidget._setXCenter(x_center)
        _stepSizeAndRangeWidget.setNCut(_nb_slices)
        if scan_type == PSI_CHAR + " angle":
            scan_type = "psi angle"
        self._tabs._outputWidget._setScanType(scan_type)

    def load(self):
        dialog = qt.QFileDialog(self)
        dialog.setNameFilters(["ini (*.ini)"])
        dialog.setFileMode(qt.QFileDialog.ExistingFile)

        if not dialog.exec_():
            dialog.close()
            return
        else:
            self.loadFile(dialog.selectedFiles()[0])

    def loadFile(self, _file):
        _logger.info(("loading", _file))
        try:
            conf = dictdump.load(_file)
        except Exception as error:
            _logger.warning(
                "Fail to load configuration from ", _file, ". Reason is", error
            )
        else:
            if self._OPT_RECO_SECTION_ID not in conf:
                _logger.warning(
                    "Dict load is not recognized. Fail to load configuration"
                )
                reco_opts = {}
            else:
                reco_opts = conf[self._OPT_RECO_SECTION_ID]
            if self._OPT_PREPROCESS_SECTION_ID not in conf:
                _logger.warning(
                    "Dict load is not recognized. Fail to load configuration"
                )
                preprocess_opts = {}
            else:
                preprocess_opts = conf[self._OPT_PREPROCESS_SECTION_ID]
            if self._TOMWER_CONFIG not in conf:
                _logger.warning(
                    "Dict load is not recognized. Fail to load configuration"
                )
                tomwer_config = {}
            else:
                tomwer_config = conf[self._TOMWER_CONFIG]
            self.setConfigFromOptionsDict(
                reco_opts=reco_opts,
                preprocess_opts=preprocess_opts,
                tomwer_config=tomwer_config,
            )

    def save(self):
        dialog = qt.QFileDialog(self)
        dialog.setNameFilters(["ini (*.ini)"])
        dialog.setAcceptMode(qt.QFileDialog.AcceptSave)
        dialog.setFileMode(qt.QFileDialog.AnyFile)

        if not dialog.exec_():
            dialog.close()
            return
        else:
            self.saveTo(dialog.selectedFiles()[0])

    def saveTo(self, _file):
        """Save the parameters used for reconstruction into an ini file"""
        if _file.lower().endswith(".ini") is False:
            _file = _file + ".ini"
        _logger.info(("saving to", _file))
        fixedOptions = self.getParameters()
        additionalRecoOptions = self.getAdditionalRecoOptions()
        additionalPreprocessOptions = self.getAdditionalPreprocessOptions()
        pre_proc_ffc = fixedOptions["ffc-when"]
        if type(pre_proc_ffc) is str:
            pre_proc_ffc = getattr(FFCWhen, pre_proc_ffc)
        assert isinstance(pre_proc_ffc, FFCWhen)

        def store(options, key, ddict):
            optionsList = _retrieve_opts_recons_cmd(
                scan_id=None,
                recons_param=fixedOptions,
                additional_opts=options,
                pre_proc_ffc=pre_proc_ffc is FFCWhen.preprocessing,
            )
            optionsList = optionsList.split("--")
            optionsDict = {}
            for opt in optionsList:
                if opt == " ":
                    continue
                _opt = opt.lstrip(" ")
                if "=" in _opt:
                    optName, value = _opt.split("=", 1)
                    value = value.replace("=", "")
                elif " " in _opt:
                    optName, value = _opt.split(" ", 1)
                    value = value.replace(" ", "")
                else:
                    optName = _opt
                    value = ""
                optName = optName.rstrip(" ")
                optionsDict[optName] = value
            ddict[key] = optionsDict

        ddict = {}
        store(options=additionalRecoOptions, key=self._OPT_RECO_SECTION_ID, ddict=ddict)
        store(
            options=additionalPreprocessOptions,
            key=self._OPT_PREPROCESS_SECTION_ID,
            ddict=ddict,
        )
        config_dict = {
            "rm-tif": fixedOptions["rm-tif"],
            "ffc-when": fixedOptions["ffc-when"],
            "half-acquisition": fixedOptions["half-acquisition"],
            "blend": fixedOptions["blend"],
            "adjust-mean": fixedOptions["adjust-mean"],
            "z": fixedOptions["z"],
        }
        ddict[self._TOMWER_CONFIG] = config_dict
        dictdump.dicttoini(ddict=ddict, inifile=_file)

    def setConfigFromOptionsDict(self, reco_opts, preprocess_opts, tomwer_config):
        """
        extract from the command line the list of option names and their
        values if any to store them

        :param list or tuple options:
        """
        self._tabs._expertWidget.resetAdditionalRecoOptions()
        self._tabs._expertWidget.resetAdditionalPreprocessOptions()
        add_options = self.setParameters(reco_opts)
        for option, optValue in add_options.items():
            self._tabs._expertWidget.addAdditionalRecoOption(
                option=option, value=optValue
            )
        add_options = self.setParameters(preprocess_opts)
        for option, optValue in add_options.items():
            self._tabs._expertWidget.addAdditionalPreprocessOption(
                option=option, value=optValue
            )
        if "rm-tif" in tomwer_config:
            self._tabs._outputWidget.setRemoveTiff(tomwer_config["rm-tif"])
        if "half-acquisition" in tomwer_config:
            self._tabs._inputWidget.setHalfAcquisition(
                tomwer_config["half-acquisition"]
            )
        if "ffc-when" in tomwer_config:
            self._tabs._inputWidget.setWhenApplyFFC(tomwer_config["ffc-when"])
        if "blend" in tomwer_config:
            self._tabs._inputWidget.setBlend(tomwer_config["blend"])
        if "adjust-mean" in tomwer_config:
            self._tabs._inputWidget.setAdjustMean(tomwer_config["adjust-mean"])
        if "z" in tomwer_config:
            self._tabs._outputWidget.setZ(tomwer_config["z"])

    def setAdditionalRecoOptions(self, options):
        assert isinstance(options, str)
        self._tabs._expertWidget.resetAdditionalRecoOptions()

    def setAdditionalPreprocessOptions(self, options):
        assert isinstance(options, str)
        self._tabs._expertWidget.resetAdditionalPreprocessOptions()

    def loadFromScan(self, scanID):
        self._tabs.loadFromScan(scanID)
        self._scanTypeChanged()


class _TofuTab(_TofuOptionLoader, qt.QTabWidget):
    """
    Tab containing the different editor for the tofu reconstruction parameter
    """

    def __init__(self, parent):
        qt.QTabWidget.__init__(self, parent)

        self._outputWidget = OutputTofuWidget(parent=self)
        icon = icons.getQIcon("output")
        self.addTab(self._outputWidget, icon, "Output definition")
        self._inputWidget = InputProjectionsScrollArea(parent=self)
        icon = icons.getQIcon("input")
        self.addTab(self._inputWidget, icon, "Projections")
        self._expertWidget = ExpertTofuWidget(parent=self)
        icon = icons.getQIcon("parameters")
        self.addTab(self._expertWidget, icon, "Expert")
        # TODO: warning, _expertWidget should always set as last tab
        _TofuOptionLoader.__init__(
            self,
            options=[],
            childs=[self._outputWidget, self._inputWidget, self._expertWidget],
        )

        # connect signal/SLOT
        self._expertWidget._gpuGrp.gpuOptionChanged.connect(self.gpuOptionUpdated)

    def loadFromScan(self, scanID):
        for widget in (self._outputWidget, self._inputWidget, self._expertWidget):
            widget.loadFromScan(scanID=scanID)

    def getAdditionalRecoOptions(self):
        return self._expertWidget.getAdditionalRecoOptions()

    def getAdditionalPreprocessOptions(self):
        return self._expertWidget.getAdditionalPreprocessOptions()

    def gpuOptionUpdated(self):
        highLimit = self._expertWidget.getHighLimit()
        self._outputWidget.setHighLimit(highLimit)


class ScanTypeWidget(_TofuOptionLoader, qt.QWidget):
    """
    Simple widget to define the type of reconstruction we want
    """

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        _TofuOptionLoader.__init__(self, options={"scan_type": self.setScanType})

        self.setLayout(qt.QHBoxLayout())
        self.layout().addWidget(qt.QLabel("scan type:", parent=self))
        self._scanType = qt.QComboBox(parent=self)
        for _type in SCAN_TYPES:
            self._scanType.addItem(_type)
        self.layout().addWidget(self._scanType)

    def setScanType(self, scan_type):
        assert scan_type in SCAN_TYPES_I
        index = self._scanType.findText(SCAN_TYPES_I[scan_type])
        assert index >= 0
        self._scanType.setCurrentIndex(index)

    def getScanType(self):
        return SCAN_TYPES[self._scanType.currentText()]



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "21/07/2020"

from silx.gui import qt
from orangewidget import widget, gui
from orangewidget.widget import Input
from tomwer.gui.visualization.dataviewer import DataViewer
from tomwer.core.scan.scanbase import TomwerScanBase
from orangewidget import settings
from tomwer.gui import utils
from tomwer.gui import icons
import tomwer.core.process.visualization.dataviewer
import logging

_logger = logging.getLogger(__name__)


class DataViewerOW(widget.OWBaseWidget, openclass=True):
    """a data viewer able to:

    - display slices (latest reconstructed if any)
    - display radios with or without normalization

    :param parent: the parent widget
    """

    name = "data viewer"
    id = "orange.widgets.tomwer.dataviewer"
    description = "allow user too browse through data"
    icon = "icons/eye.png"
    priority = 70
    keywords = ["tomography", "file", "tomwer", "acquisition", "validation"]

    want_main_area = True
    resizing_enabled = True
    compress_signal = False

    ewokstaskclass = tomwer.core.process.visualization.dataviewer._DataViewerPlaceHolder

    _viewer_config = settings.Setting(dict())

    class Inputs:
        data = Input(name="data", type=TomwerScanBase)

    def __init__(self, parent=None):
        widget.OWBaseWidget.__init__(self, parent)
        self._layout = gui.vBox(self.mainArea, self.name).layout()
        self.viewer = DataViewer(parent=self)
        self.viewer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Expanding)
        self._layout.addWidget(self.viewer)
        self._setSettings(settings=self._viewer_config)

        # open with ImageJ button
        types = qt.QDialogButtonBox.Apply
        self._buttons = qt.QDialogButtonBox(self)
        self._buttons.setStandardButtons(types)
        self._buttonOpenImageJ = self._buttons.button(qt.QDialogButtonBox.Apply)
        self._buttonOpenImageJ.setText("open with ImageJ")
        self._buttonOpenImageJ.setIcon(icons.getQIcon("Imagej_icon"))
        self.layout().addWidget(self._buttons)

        # connect signal / slot
        self.viewer.sigConfigChanged.connect(self._updateSettings)
        self._buttonOpenImageJ.released.connect(self._openCurrentInImagej)

    def _openCurrentInImagej(self):
        current_url = self.viewer.getCurrentUrl()

        if current_url is None:
            _logger.warning("No active image defined")
        else:
            try:
                utils.open_url_with_image_j(current_url)
            except OSError as e:
                msg = qt.QMessageBox(self)
                msg.setIcon(qt.QMessageBox.Warning)
                msg.setWindowTitle("Unable to open image in imagej")
                msg.setText(str(e))
                msg.exec_()

    @Inputs.data
    def addScan(self, scan):
        if scan is None:
            return
        self.viewer.setScan(scan)

    def sizeHint(self):
        return qt.QSize(400, 500)

    def _updateSettings(self):
        self._viewer_config["mode"] = self.viewer.getDisplayMode()
        self._viewer_config["slice_opt"] = self.viewer.getSliceOption()
        self._viewer_config["radio_opt"] = self.viewer.getRadioOption()

    def _setSettings(self, settings):
        old_state = self.viewer.blockSignals(True)
        if "mode" in settings:
            self.viewer.setDisplayMode(settings["mode"])
        if "slice_opt" in settings:
            self.viewer.setSliceOption(settings["slice_opt"])
        if "radio_opt" in settings:
            self.viewer.setRadioOption(settings["radio_opt"])
        self.viewer.blockSignals(old_state)

    def close(self):
        self.viewer.close()
        self.viewer = None
        super().close()



# /*##########################################################################
# Copyright (C) 2016-2017 European Synchrotron Radiation Facility
#
# This file is part of the PyMca X-ray Fluorescence Toolkit developed at
# the ESRF by the Software group.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__author__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "06/08/2018"


import functools
import logging
import os

from silx.gui import qt
from silx.io.url import DataUrl

from tomwer.core.utils import ftseriesutils
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.synctools.imageloaderthread import ImageLoaderThread
from tomoscan.io import HDF5File

logger = logging.getLogger(__name__)


class _Image(qt.QObject):
    def __init__(self, data):
        qt.QObject.__init__(self)
        self.data = data


class ImageFromFile(_Image):
    """
    Define an Image with a status
    """

    LOADING_STATUS = {"loading": 0, "loaded": 1, "not loaded": 2}
    sigLoaded = qt.Signal(str)
    """Signal emitted when the file is loaded. Parameter is the file url"""

    def __init__(self, _file=None, index=None, data=None, _load=False, url=None):
        _Image.__init__(self, data)
        if url is not None:
            assert _file is None and index is None
            self.url = url
        else:
            assert isinstance(_file, str)
            data_path = None
            if _file.lower().endswith(".edf"):
                scheme = "fabio"
            elif _file.lower().endswith((".npy", ".npz")):
                scheme = "numpy"
            else:

                def get_nabu_entry():
                    try:
                        with HDF5File(_file, mode="r", swmr=True) as h5s:
                            for node in h5s:
                                if "reconstruction" in h5s[node]:
                                    return "/".join(
                                        (node, "reconstruction", "results", "data")
                                    )
                    except Exception as e:
                        logger.error(e)

                data_path = get_nabu_entry() or None
                scheme = "silx"

            if index is not None:
                data_slice = (index,)
            else:
                data_slice = None
            self.url = DataUrl(
                file_path=_file,
                data_slice=data_slice,
                scheme=scheme,
                data_path=data_path,
            )

        self._status = "not loaded" if data is None else "loaded"
        self.url_path = self.url.path()
        if _load is True:
            self.load()

    def load(self, sync=False):
        """
        Load the data contained in the url.

        :param sync: if True: then wait for the image to be loaded for giving
            back hand
        """
        if self._status == "loaded" and self.data is not None:
            self.sigLoaded.emit(self.url.path())
        elif self._status == "loading":
            logger.debug("%s is already loading" % self.url.path())
        else:
            self._status = "loading"
            self.loaderThread = ImageLoaderThread(url=self.url)

            callback = functools.partial(
                self._setData, self.url_path, self.loaderThread.getData
            )
            self.loaderThread.finished.connect(callback)
            self.loaderThread.start()
            if sync is False:
                self.loaderThread.wait()

    def _setData(self, url_path, dataGetter):
        self.data = dataGetter()
        if self.data is not None:
            if self.data.ndim != 2:
                if self.data.shape[0] == 1:
                    self.data = self.data.reshape(self.data.shape[1:])
                elif self.data.shape[-1] == 1:
                    self.data = self.data.reshape(self.data.shape[0:-1])

        self._status = "loaded"
        self.sigLoaded.emit(url_path)

    def isLoaded(self):
        return self._status == "loaded"


class FileWithImage(object):
    """Definition of file which can contain multiple images"""

    def __init__(self, _file):
        self.file = _file

    def getImages(self, _load):
        """

        :param bool _load: if True then launch the load of the image
        :return: the list of images contained in the given file.
        """
        if self.file.endswith(".vol"):
            return self._dealWithVolFile(self.file, _load)
        elif self.file.lower().endswith(TomwerScanBase.VALID_RECONS_EXTENSION):
            return [ImageFromFile(self.file)]
        else:
            logger.error("only deal with .edf and .vol extension")
            return []

    def _dealWithVolFile(self, _file, _load):
        _imagesFiles = []
        volInfoFile = _file.replace(".vol", ".vol.info")
        if not os.path.exists(volInfoFile):
            mess = (
                "Can't find description file %s associated with raw "
                "data file %s " % (volInfoFile, _file)
            )
            logger.warning(mess)
        else:
            shape = ftseriesutils.get_vol_file_shape(volInfoFile)
            if shape is not None:
                for zSlice in range(shape[0]):
                    if _file.lower().endswith(".edf"):
                        scheme = "fabio"
                    else:
                        scheme = "tomwer"
                    _imagesFiles.append(
                        ImageFromFile(
                            url=DataUrl(
                                file_path=_file, data_slice=(zSlice,), scheme=scheme
                            ),
                            _load=_load,
                        )
                    )
            return _imagesFiles



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

"""
This module is used to define the process of the reference creator.
This is related to the issue #184
"""

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "17/08/2021"


import os
from silx.io.utils import h5py_read_dataset
import fabio
import numpy
import typing
from silx.io.url import DataUrl
from tomwer.core import utils
from silx.io.dictdump import dicttoh5, h5todict
from tomoscan.io import HDF5File
import logging
from tomwer.core.process.reconstruction.darkref.darkrefs import DarkRefs
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomwer.core.scan.edfscan import EDFTomoScan
from .settings import DARKHST_PREFIX, REFHST_PREFIX

logger = logging.getLogger(__name__)


class DarkRefsCopy(DarkRefs):
    """
    Reimplement Dark ref to deal with copy when there is no median/mean files
    """

    DEFAULT_SRCURRENT = 200.0  # mA

    def __init__(self, process_id=None, varinfo=None, inputs=None, label=None):
        super().__init__(
            process_id=process_id, varinfo=varinfo, inputs=inputs, label=label
        )
        if inputs is None:
            inputs = {}
        self._refHstPrefix = inputs.get("refhst_prefix", REFHST_PREFIX)
        self._darkHstPrefix = inputs.get("darkhst_prefix", DARKHST_PREFIX)
        self._mode_auto = inputs.get("mode_auto", True)
        if "savedir" in inputs:
            raise KeyError("savedir is not a valid key. Use save_dir")
        self._savedir = inputs.get("save_dir", None)
        if self._savedir is None:
            raise KeyError(
                "save_dir is expected (where the ref and dark file should be saved)"
            )
        self._dark_save_file = os.path.join(self._savedir, "dark.h5")
        self._ref_save_file = os.path.join(self._savedir, "refs.h5")
        info = "Ref copy will store files in " + self._savedir
        logger.info(info)
        """directory saved to store the dark and refHST reference files"""
        self._processOnlyCopy = inputs.get("process_only_copy", False)
        self._processOnlyDkRf = inputs.get("process_only_dkrf", False)
        init_ref_scan = inputs.get("init_dkrf_from", None)
        if init_ref_scan:
            if not isinstance(init_ref_scan, TomwerScanBase):
                raise TypeError(
                    "init_ref_scan is expected to be an instance of TomwerScanBase. Not {}".format(
                        type(init_ref_scan)
                    )
                )
            else:
                self.set_refs_from_scan(init_ref_scan)

    def set_process_only_dkRf(self, value: bool) -> None:
        self._processOnlyDkRf = value
        self._processOnlyCopy = False

    def set_process_only_copy(self, value: bool) -> None:
        self._processOnlyDkRf = False
        self._processOnlyCopy = value

    def set_refs_from_scan(self, scan: TomwerScanBase) -> bool:
        assert isinstance(scan, TomwerScanBase)
        if isinstance(scan, HDF5TomoScan):
            darks = DarkRefs.get_darks_frm_process_file(
                scan.process_file, entry=scan.entry, as_url=True
            )
            if len(darks) == 0:
                logger.warning("No dark found in %s")
            flats = DarkRefs.get_flats_frm_process_file(
                scan.process_file, entry=scan.entry, as_url=True
            )
            if len(flats) == 0:
                logger.warning("No flat found in %s")
            has_flat_or_dark = len(darks) > 0 or len(flats) > 0
            if not has_flat_or_dark:
                self._notice_user_no_ref_message(fileType="flat field", dir=scan)
                return False
            else:
                self.clean_save_files()
                if len(darks) == 0:
                    self._notice_user_no_ref_message(fileType="dark", dir=scan)
                else:
                    key_0 = sorted(darks.keys())[0]
                    self._copy_dark_from(dark_url=darks[key_0], scan=scan)

                if len(flats) == 0:
                    self._notice_user_no_ref_message(fileType="flat field", dir=scan)
                else:
                    key_0 = sorted(flats.keys())[0]
                    self._copy_ref_from(url=flats[key_0], scan=scan)
                self._ref_has_been_set(scan=scan)
                return True
        elif isinstance(scan, EDFTomoScan):
            if os.path.isdir(scan.path) is False:
                w = "given path (%s) is not a directory. Can't extract ref" % scan
                logger.warning(w)
                return False

            self.directory = scan
            if not self.contains_dark_or_ref(scan):
                self._notice_user_no_ref_message(fileType="flat field", dir=scan)
                return False
            else:
                self.clean_save_files()
                darkHSTFiles = sorted(
                    DarkRefs.getDarkHSTFiles(scan.path, prefix=self.darkHST_prefix)
                )
                refHSTFiles = sorted(
                    DarkRefs.getRefHSTFiles(scan.path, prefix=self.refHST_prefix)
                )
                if len(darkHSTFiles) == 0:
                    self._notice_user_no_ref_message(fileType="dark", dir=scan)
                else:
                    self._copy_dark_from(
                        dark_url=DataUrl(file_path=darkHSTFiles[0], scheme="fabio"),
                        scan=scan,
                    )

                if len(refHSTFiles) == 0:
                    self._notice_user_no_ref_message(fileType="flat field", dir=scan)
                else:
                    self._copy_ref_from(
                        url=DataUrl(file_path=refHSTFiles[0], scheme="fabio"), scan=scan
                    )

                """warning: dark should always be saved first because used for
                saving ref
                """
                self._ref_has_been_set(scan=scan)
                return True

    def _ref_has_been_set(self, scan):
        # TODO: this should be removed and set in the QObject inheritingtest_axis_darkrefs_ftseries from the DkrfCopy
        if hasattr(self, "sigRefSetted"):
            self.sigRefSetted.emit(scan)

    def _copy_to(self, scan):
        if not isinstance(scan, TomwerScanBase):
            raise TypeError(
                "{} is expected to be a {} and not {}".format(
                    str(scan), TomwerScanBase, type(scan)
                )
            )

        if isinstance(scan, EDFTomoScan):
            if os.path.isdir(scan.path) is False:
                logger.warning(
                    str(scan) + " is not a directory. Cannot copy " "reference to it."
                )
                return
            if self.has_flat_or_dark_stored() is False:
                logger.error("No reference registred to be copy in %s" % scan)
                return

        # warning: get information if there is ref or dark before copying
        # because can bring interferences.
        _hasRef = self.contains_flat(scan)
        _hasDark = self.contains_dark(scan)
        # special treatment for dark and ref we want to store both at the same
        # time
        if isinstance(scan, HDF5TomoScan) and not _hasDark and not _hasRef:
            self._copy_dark_and_flat_to(scan)
        else:
            if self.has_dark_stored() and not _hasDark:
                self._copy_dark_to(scan)
            if self.has_flat_stored() and not _hasRef:
                self._copy_flat_to(scan)

    def _copy_dark_and_flat_to(
        self, scan: HDF5TomoScan, copy_dark: bool = True, copy_flats: bool = True
    ):
        """
        Function dedicated to the copy of dark and/or flat to HDF5TomoScan
        :param HDF5TomoScan scan:
        :param bool copy_dark:
        :param bool copy_flats:
        :return:
        """
        assert isinstance(scan, HDF5TomoScan), "function dedicated to HDF5TomoScan"
        if copy_dark is False and copy_flats is False:
            return

        results = {}
        shape_dark = None
        shape_flat = None
        if copy_dark:
            dark_info = self._get_dark_info()
            if dark_info is not None:
                results["darks"] = {"0": dark_info["data"]}
                shape_dark = dark_info["data"].shape
                scan.set_normed_darks(results["darks"])
        if copy_flats:
            flat_info = self._get_flat_info()
            if flat_info is not None:
                flats = {"0": flat_info["data_start"]}
                shape_flat = flat_info["data_start"].shape
                index_last_projection = max(scan.projections) or None
                if index_last_projection is not None and index_last_projection != 0:
                    flats[str(index_last_projection)] = flat_info["data_end"]
                results["flats"] = flats
                scan.set_normed_flats(results["flats"])

        # check dimensions of dark and flat compare to projections if any
        if scan.dim_1 is not None and scan.dim_2 is not None:
            shape_frame = (scan.dim_2, scan.dim_1)
            incoherent_with_dark = False
            incoherent_with_flat = False
            if shape_dark is not None:
                incoherent_with_dark = shape_dark != shape_frame
            if shape_flat is not None:
                incoherent_with_flat = shape_flat != shape_frame

            if incoherent_with_dark:
                logger.warning(
                    "Scan frame (%s) and dark (%s) to copy have"
                    "different shapes. Unable to do the copy"
                    "" % (shape_frame, shape_dark)
                )
            if incoherent_with_flat:
                logger.warning(
                    "Scan frame (%s) and flat (%s) to copy have"
                    "different shapes. Unable to do the copy"
                    "" % (shape_frame, shape_flat)
                )
            if incoherent_with_flat or incoherent_with_dark:
                return

        try:
            with scan.acquire_process_file_lock():
                DarkRefs._register_process(
                    process_file=scan.process_file,
                    entry=scan.entry,
                    process=DarkRefs,
                    configuration=self.recons_params.to_dict(),
                    results=results,
                    process_index=scan.pop_process_index(),
                    overwrite=True,
                )
        except Exception as e:
            logger.warning(f"Fail to register darkrefcopy process. Reason is {e}")

    def _copy_dark_to(self, scan):
        """The dark is copied without normalization"""
        assert isinstance(scan, TomwerScanBase)
        if not self.has_dark_stored():
            raise ValueError("require to copy dark but no dark stored")
        dark_info = self._get_dark_info()
        assert dark_info is not None, (
            "when this function is call you should insure"
            "that process has stored some dark"
        )
        if isinstance(scan, EDFTomoScan):
            # do never copy on an existing file (meaning ref already here)
            dark_file_name = dark_info["dark_basename"] or DARKHST_PREFIX
            dst = os.path.join(scan.path, os.path.basename(dark_file_name))
            if os.path.isfile(dst) is True:
                logger.info("%s exists, will not copy dark" % dst)
            else:
                header = dark_info["header"]
                data = dark_info["data"]
                file_desc = fabio.edfimage.EdfImage(data=data, header=header)
                file_desc.write(dst, force_type=numpy.int32)

        elif isinstance(scan, HDF5TomoScan):
            self._copy_dark_and_flat_to(scan=scan, copy_dark=True, copy_flats=False)
        else:
            raise ValueError("Tomo scan type not managed (%s)" % scan)

    def _copy_flat_to(self, scan):
        assert isinstance(scan, TomwerScanBase)
        if not self.has_flat_stored():
            raise ValueError("require to copy flat but no flat stored")
        if isinstance(scan, EDFTomoScan):

            def normalize(data, when):
                """normalize from dark_end and getting the SRCurrent at start or
                end"""
                assert when in ("start", "end")

                srCurrent = utils.getSRCurrent(scan_dir=scan.path, when=when)
                if srCurrent is None:
                    logger.warning(
                        "Can't find information about srCurrent," "set to default value"
                    )
                    srCurrent = self.DEFAULT_SRCURRENT
                dark_info = self._get_dark_info()
                if dark_info is None:
                    logger.info("no dark found, unable to normalize")
                    return None, srCurrent
                dark_data = dark_info["data"]
                if data.shape != dark_data.shape:
                    logger.warning(
                        "Image and dark have different shapes."
                        "Cannot normalize. (Dark ref file in %s)"
                        "" % self._dark_save_file
                    )
                    return data, srCurrent
                return (data * srCurrent).astype(numpy.float32) + dark_data, srCurrent

            flat_info = self._get_flat_info()
            if flat_info is None:
                raise RuntimeError("No ref found during processing")
            tomo_N = utils.getTomo_N(scan.path)
            if tomo_N in (None, -1):
                logger.error(
                    "Can't find the number of projection. " "Fail to create reference"
                )

            header = flat_info["header"]
            end_acqui = str(utils.getTomo_N(scan.path)).zfill(4)
            indexRefFile = {"start": "0000", "end": end_acqui}
            for when in ("start", "end"):
                data = flat_info["_".join(("data", when))]
                fileName = self.recons_params.ref_prefix + indexRefFile[when] + ".edf"
                filePath = os.path.join(scan.path, fileName)
                # do never copy on an existing file (meaning ref already here)
                if os.path.isfile(filePath):
                    continue
                _data, _srCurrent = normalize(data, when)

                _header = header.copy()
                # add some extra information on the header
                _header["SRCUR"] = _srCurrent
                file_desc = fabio.edfimage.EdfImage(data=_data, header=_header)
                file_desc.write(filePath, force_type=numpy.int32)
        elif isinstance(scan, HDF5TomoScan):
            self._copy_dark_and_flat_to(scan=scan, copy_dark=False, copy_flats=True)
        else:
            raise ValueError("Tomo scan type not managed (%s)" % scan)

    def _notice_user_no_ref_message(self, fileType, dir):
        logger.warning(self._get_no_ref_message(fileType, dir))

    def _get_no_ref_message(self, fileType, dir):
        mess = (
            "No %s found in the given directory %s. Won't be able to copy"
            " them" % (fileType, dir)
        )
        return mess

    def _copy_ref_from(self, scan, url):
        """
        Copy the data contained in _file in the given file to the `_savedir`
        and normalize the data from SRCurrent (intensity
        )"""

        def write(data_start, data_end, header):
            with HDF5File(self._ref_save_file, "w") as h5f:
                h5f["data_start"] = data_start
                h5f["data_end"] = data_end
                h5f["tomwer_info"] = "copied using refCopy"
                h5f["original_scan"] = str(scan)
                h5f["srcurrent"] = -1
            # update header if any edf write
            header["tomwer_info"] = "copied using refCopy"
            header["original_scan"] = str(scan)
            header["SRCUR"] = -1
            dicttoh5(
                header,
                h5file=self._ref_save_file,
                h5path="header",
                overwrite_data=False,
                mode="a",
            )

        assert isinstance(url, DataUrl)
        if url.scheme() == "fabio":

            def normalize_edf(data):
                srCurrent = utils.getClosestSRCurrent(
                    scan_dir=scan.path, refFile=url.file_path()
                )
                if srCurrent in (None, -1):
                    logger.warning(
                        "Can't find information about srCurrent,"
                        "set to default value for normalization"
                    )
                    srCurrent = self.DEFAULT_SRCURRENT

                if not self.contains_dark(scan=scan):
                    logger.warning("No darkHST recorded, unable to normalize")
                    return data, srCurrent

                dark_data = self._get_dark_info()
                if dark_data is None:
                    logger.warning("No darkHST found in %s" % self._dark_save_file)
                    return None, None
                else:
                    dark_data = dark_data["data"]
                if data.shape != dark_data.shape:
                    err = (
                        "cannot normalize data from %s, has different "
                        "dimensions" % (self._darkHST)
                    )
                    logger.error(err)
                    return data, srCurrent
                return (data - dark_data).astype(numpy.float32) / srCurrent, srCurrent

            with fabio.open(url.file_path()) as file_desc:
                data = file_desc.data
                header = file_desc.header
                data, sr_current = normalize_edf(data)
                write(data_start=data, data_end=data, header=header)
                # TODO: look but the writing should be more or less generic
        elif url.scheme() in ("silx", "h5py"):
            flats = DarkRefs.get_flats_frm_process_file(
                scan.process_file, entry=scan.entry, as_url=False
            )
            if len(flats) == 0:
                logger.warning(
                    "No flat found in %s, unable to copy" % scan.process_file
                )
                return
            key_0 = sorted(list(flats.keys()))[0]
            data_start = flats[key_0]
            if len(flats) > 2:
                key_end = sorted(list(flats.keys()))[-1]
                data_end = flats[key_end]
            else:
                data_end = data_start
            header = {}
            write(data_start=data_start, data_end=data_end, header=header)
        else:
            raise ValueError("scheme not managed")

    def _copy_dark_from(self, scan, dark_url: DataUrl):
        assert isinstance(dark_url, DataUrl)
        assert isinstance(scan, TomwerScanBase)
        if dark_url.scheme() == "fabio":
            with fabio.open(dark_url.file_path()) as dsc:
                header = dsc.header.copy()
                data = dsc.data.copy()
                dark_file_basename = os.path.basename(dark_url.file_path())
        elif dark_url.scheme() in ("h5py", "silx"):
            darks = DarkRefs.get_darks_frm_process_file(
                scan.process_file, entry=scan.entry, as_url=False
            )
            if len(darks) == 0:
                logger.warning(
                    "No dark found in %s, unable to copy" % scan.process_file
                )
                return
            key_0 = sorted(list(darks.keys()))[0]
            data = darks[key_0]
            dark_file_basename = None
            # TODO: get some metadata ?
            header = {}
        else:
            raise ValueError("scheme not managed")

        # add origin of the dark file
        with HDF5File(self._dark_save_file, "w") as h5f:
            # h5f['header'] = header
            h5f["tomwer_info"] = "copied using refCopy"
            h5f["original_scan"] = str(scan)
            h5f["data"] = data
            h5f["srcurrent"] = -1
            if dark_file_basename is not None:
                h5f["dark_basename"] = dark_file_basename
        header["tomwer_info"] = "copied using refCopy"
        header["original_scan"] = str(scan)
        header["SRCUR"] = -1
        dicttoh5(
            header,
            h5file=self._dark_save_file,
            h5path="header",
            overwrite_data=False,
            mode="a",
        )

    def _get_flat_info(self) -> typing.Union[None, dict]:
        """return ref data. Only one store for now"""
        if self.has_flat_stored():
            with HDF5File(self._ref_save_file, "r", swmr=True) as h5f:
                return {
                    "data_start": h5py_read_dataset(h5f["data_start"]),
                    "data_end": h5py_read_dataset(h5f["data_end"]),
                    "original_scan": h5py_read_dataset(h5f["original_scan"]),
                    "srcurrent": h5py_read_dataset(h5f["srcurrent"]),
                    "header": h5todict(h5file=self._dark_save_file, path="header"),
                    "tomwer_info": h5py_read_dataset(h5f["tomwer_info"]),
                }
        else:
            return None

    def _get_dark_info(self) -> typing.Union[None, dict]:
        """return dark data, header, original_scan"""
        if self.has_dark_stored():
            with HDF5File(self._dark_save_file, "r", swmr=True) as h5f:
                if "dark_basename" in h5f:
                    dark_basename = h5py_read_dataset(h5f["dark_basename"])
                else:
                    dark_basename = None
                res = {
                    "data": h5py_read_dataset(h5f["data"]),
                    "original_scan": h5py_read_dataset(h5f["original_scan"]),
                    "srcurrent": h5py_read_dataset(h5f["srcurrent"]),
                    "dark_basename": dark_basename,
                    "tomwer_info": h5py_read_dataset(h5f["tomwer_info"]),
                    "header": h5todict(h5file=self._dark_save_file, path="header"),
                }
                return res
        else:
            return None

    def clean_save_files(self):
        for f in os.listdir(self._savedir):
            if os.path.isfile(f) and os.path.exists(f):
                os.remove(f)

    def run(self):
        """
        This is function triggered when a new scan / data is received.
        As explained in issue #184 the behavior is the following:

        * if the scan has already ref files files won't be overwrite
        * if the mode is in `auto` will register last ref file met
        * if the scan has no ref files and refCopy has some register. Will
          create a copy of those, normalized from srCurrent (for flat field)
        """
        scan = self.inputs.data
        if not isinstance(scan, (type(None), TomwerScanBase, dict)):
            raise TypeError(
                f"self.inputs.data is expected to be an instance "
                f"of None, TomwerScanBase or dict. Not {type(scan)}"
            )
        if scan is None or scan.path is None:
            return
        # process dark and ref calculation
        super().run()
        # then copy if necessary
        if not self._processOnlyDkRf:
            if self.contains_dark_or_ref(scan):
                if self.is_on_mode_auto:
                    self.set_refs_from_scan(scan)
            if self.has_missing_dark_or_ref(scan):
                self._copy_to(scan)

    def has_flat_or_dark_stored(self) -> bool:
        """

        :return: True if the process has at least registered one flat or one
                 dark
        :rtype: bool
        """
        return self.has_flat_stored() or self.has_dark_stored()

    def has_flat_stored(self) -> bool:
        """

        :return: True if the process has registered at least one ref
        :rtype: bool
        """
        return os.path.exists(self._ref_save_file)

    def has_dark_stored(self) -> bool:
        """

        :return: True if the process has registered at least one dark
        :rtype: bool
        """
        return os.path.exists(self._dark_save_file)

    def contains_dark(self, scan: TomwerScanBase) -> bool:
        """Return True if the scan has already some dark processed"""
        if not isinstance(scan, TomwerScanBase):
            return TypeError(
                "scan is expected to be an instance of {} and not {}".format(
                    TomwerScanBase, type(scan)
                )
            )

        if isinstance(scan, EDFTomoScan):
            return (
                len(
                    DarkRefs.getDarkHSTFiles(
                        directory=scan.path, prefix=self.darkHST_prefix
                    )
                )
                > 0
            )
        elif isinstance(scan, HDF5TomoScan):
            if scan.process_file is None or not os.path.exists(scan.process_file):
                return False
            else:
                return (
                    len(
                        DarkRefs.get_darks_frm_process_file(
                            scan.process_file, entry=scan.entry, as_url=True
                        )
                    )
                    > 0
                )
        else:
            raise ValueError("scan type not managed")

    def contains_flat(self, scan: TomwerScanBase):
        """Return True if the scan has already some dark processed"""
        if not isinstance(scan, TomwerScanBase):
            return TypeError(
                "scan is expected to be an instance of {} and not {}".format(
                    TomwerScanBase, type(scan)
                )
            )

        if isinstance(scan, EDFTomoScan):
            return (
                len(
                    DarkRefs.getRefHSTFiles(
                        directory=scan.path, prefix=self.refHST_prefix
                    )
                )
                > 0
            )
        elif isinstance(scan, HDF5TomoScan):
            if scan.process_file is None or not os.path.exists(scan.process_file):
                return False
            else:
                return (
                    len(
                        DarkRefs.get_flats_frm_process_file(
                            scan.process_file, entry=scan.entry, as_url=True
                        )
                    )
                    > 0
                )
        else:
            raise ValueError("scan type not managed")

    def contains_dark_or_ref(self, scan):
        return self.contains_dark(scan=scan) or self.contains_flat(scan=scan)

    def has_missing_dark_or_ref(self, scan: TomwerScanBase) -> bool:
        """return True if the scan has no ref or no dark registered"""
        assert isinstance(scan, TomwerScanBase)
        return not self.contains_dark(scan) or not self.contains_flat(scan)

    def _signal_done(self, scan):
        assert isinstance(scan, TomwerScanBase)
        raise NotImplementedError("Abstract class")

    def set_mode_auto(self, b):
        self._mode_auto = b

    @property
    def is_on_mode_auto(self):
        return self._mode_auto

    @property
    def refHST_prefix(self):
        return self._refHstPrefix

    @property
    def darkHST_prefix(self):
        return self._darkHstPrefix

    def set_refHST_prefix(self, prefix):
        self._refHstPrefix = prefix

    def set_darkHST_prefix(self, prefix):
        self._darkHstPrefix = prefix

    def clear_ref(self):
        if self._ref_save_file is not None:
            os.remove(self._ref_save_file)
        if self._dark_save_file is not None:
            os.remove(self._dark_save_file)

    def get_dark_save_file(self):
        return self._dark_save_file

    def get_ref_save_file(self):
        return self._ref_save_file



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "24/01/2017"

import logging
import shutil
import tempfile
import unittest

from silx.gui import qt

from orangecontrib.tomwer.widgets.visualization.ImageStackViewerOW import (
    ImageStackViewerOW,
)
from tomwer.core.utils.scanutils import MockEDF
from tomwer.gui.qtapplicationmanager import QApplicationManager
from tomwer.core.scan.edfscan import EDFTomoScan

logging.disable(logging.INFO)

# TODO: use a QtCase instead
# Makes sure a QApplication exists
_qapp = QApplicationManager()


class TestFilePatterns(unittest.TestCase):
    """Make sure the viewer recognize each possible file pattern"""

    N_ACQUI = 20
    """number of acquisition runned"""

    N_RECONS = 4
    """Number of non Paganin reconstruction"""

    N_PAG_RECONS = 2
    """Number of Paganin reconstructions"""

    @classmethod
    def setUpClass(cls):
        cls.stackViewer = ImageStackViewerOW()
        cls.stackViewer.setAttribute(qt.Qt.WA_DeleteOnClose)

    @classmethod
    def tearDownClass(cls):
        cls.stackViewer.close()
        del cls.stackViewer

    def setUp(self):
        self._folder = tempfile.mkdtemp()
        MockEDF.fastMockAcquisition(self._folder, n_radio=self.N_ACQUI)

    def tearDown(self):
        shutil.rmtree(self._folder)

    def testAcquisition(self):
        """
        Make sure the viewer is able to found and load the acquisition files
        """
        self.stackViewer.addScan(EDFTomoScan(self._folder))

        radioDict = self.stackViewer.viewer._stackImageViewerRadio.images
        self.assertTrue(radioDict.size() == self.N_ACQUI)

    def testReconsNotPag(self):
        """
        Make sure the viewer is able to found and load the reconstruction files
        """
        MockEDF.mockReconstruction(self._folder, nRecons=self.N_RECONS)
        self.stackViewer.addScan(EDFTomoScan(self._folder))

        scanDict = self.stackViewer.viewer._stackImageViewerScan.images
        self.assertTrue(scanDict.size() == self.N_RECONS)

    def testReconsPaganin(self):
        """
        Make sure the viewer is able to found and load the paganin recons files
        """
        MockEDF.mockReconstruction(
            self._folder, nRecons=0, nPagRecons=self.N_PAG_RECONS
        )
        self.stackViewer.addScan(EDFTomoScan(self._folder))

        scanDict = self.stackViewer.viewer._stackImageViewerScan.images
        self.assertTrue(scanDict.size() == self.N_PAG_RECONS)

    def testAllRecons(self):
        """
        Make sure the viewer is able to found and load the paganin and the
        not paganin reconstruction and display both
        """
        MockEDF.mockReconstruction(
            self._folder, nRecons=self.N_RECONS, nPagRecons=self.N_PAG_RECONS
        )
        self.stackViewer.addScan(EDFTomoScan(self._folder))

        scanDict = self.stackViewer.viewer._stackImageViewerScan.images
        self.assertTrue(scanDict.size() == self.N_RECONS + self.N_PAG_RECONS)

    def testReconsVolFile(self):
        """Make sure the viewer is able to detect .vol file"""
        MockEDF.mockReconstruction(
            self._folder, nRecons=self.N_RECONS, nPagRecons=0, volFile=True
        )
        self.stackViewer.addScan(EDFTomoScan(self._folder))
        scanDict = self.stackViewer.viewer._stackImageViewerScan.images
        # we create only one .vol file
        self.assertEqual(scanDict.size(), self.N_RECONS + 1)



# /*##########################################################################
# Copyright (C) 2016-2017 European Synchrotron Radiation Facility
#
# This file is part of the PyMca X-ray Fluorescence Toolkit developed at
# the ESRF by the Software group.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__author__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "10/01/2018"


import logging
from silx.gui import qt
from tomwer.gui.reconstruction.ftserie.h5editor import H5StructEditor
from tomwer.synctools.ftseries import QReconsParams, _QFTRP

logger = logging.getLogger(__name__)


class ExpertWidget(H5StructEditor, qt.QWidget):
    """
    Create the widget inside the Expert tab of the SimplifyH5ParamEditor

    :param reconsparams: reconstruction parameters edited by the widget
    """

    def __init__(self, reconsparams, parent=None):
        qt.QWidget.__init__(self, parent)
        H5StructEditor.__init__(self, structID="FT")
        self._recons_params = None
        self.setReconsParams(recons_params=reconsparams)

        self.setLayout(qt.QVBoxLayout())
        self.layout().addWidget(self.__buildNumPart())
        self.layout().addWidget(self.__buildVersion())
        self.layout().addWidget(self.__buildDataBase())
        self.layout().addWidget(self.__buildNoCheck())
        self.layout().addWidget(self.__buildZeroOffMask())
        self.layout().addWidget(self.__buildFixHD())
        # spacer
        self.spacer = qt.QWidget(self)
        self.spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(self.spacer)

        self._makeConnection()

    def setReconsParams(self, recons_params):
        # TODO: update gui when setting reconsparam to
        if isinstance(recons_params, QReconsParams):
            _recons_params = recons_params.ft
        elif isinstance(recons_params, _QFTRP):
            _recons_params = recons_params
        else:
            raise ValueError(
                "recons_params should be an instance of QReconsParam or _QFTRP"
            )

        if self._recons_params:
            self._recons_params.sigChanged.disconnect(self._update_params)
        self._recons_params = _recons_params
        self.load(self._recons_params)
        self._recons_params.sigChanged.connect(self._update_params)

    def _update_params(self):
        """Update all parameter"""
        assert isinstance(self._recons_params, _QFTRP)
        self.load(self._recons_params)

    def _makeConnection(self):
        self._qsbNumPart.valueChanged.connect(self._numPartChanged)
        self._qcbDataBase.toggled.connect(self._dataBaseChanged)
        self._qcbNocheck.toggled.connect(self._noCheckChanged)
        self._qcbZeroRegionMask.toggled.connect(self._zeroRegionMaskChanged)
        self._qcbFixHD.toggled.connect(self._fixHDChanged)

    def __buildNumPart(self):
        widget = qt.QWidget(self)
        widget.setLayout(qt.QHBoxLayout())
        widget.layout().addWidget(
            qt.QLabel("length of the numerical part in the data filenames ")
        )

        self._qsbNumPart = qt.QSpinBox(parent=widget)
        self._qsbNumPart.setMinimum(0)
        widget.layout().addWidget(self._qsbNumPart)
        self.linkGroupWithH5Variable(
            group=self._qsbNumPart,
            h5ParamName="NUM_PART",
            setter=self._setNumericalPart,
            getter=self._getNumericalPart,
        )
        return widget

    def _numPartChanged(self):
        if self._isLoading is False:
            value = self._qsbNumPart.value()
            self._recons_params._set_parameter_value(parameter="NUM_PART", value=value)

    def __buildVersion(self):
        widget = qt.QWidget(self)
        widget.setLayout(qt.QHBoxLayout())
        widget.layout().addWidget(qt.QLabel("Version used:", parent=widget))

        self._qleVersion = qt.QLabel("", widget)
        widget.layout().addWidget(self._qleVersion)
        self.linkGroupWithH5Variable(
            self._qleVersion,
            "VERSION",
            getter=self._qleVersion.text,
            setter=self._qleVersion.setText,
        )

        spacer = qt.QWidget(widget)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        widget.layout().addWidget(spacer)
        # hide this widget, since fastomo 3 is not used
        widget.hide()

        return widget

    def __buildDataBase(self):
        self._qcbDataBase = qt.QCheckBox("put scan in tomoDB", parent=self)
        self.linkCheckboxWithH5Variable(self._qcbDataBase, "DATABASE")
        # for now the tomoDB option is deactivate. Not managed for the
        # moment
        self._qcbDataBase.hide()
        return self._qcbDataBase

    def _dataBaseChanged(self, b):
        if self._isLoading is False:
            self._recons_params._set_parameter_value(parameter="DATABASE", value=int(b))

    def __buildNoCheck(self):
        self._qcbNocheck = qt.QCheckBox(
            "force reconstruction of slices in ftseries", parent=self
        )
        self.linkCheckboxWithH5Variable(self._qcbNocheck, "NO_CHECK")
        return self._qcbNocheck

    def _noCheckChanged(self, b):
        if self._isLoading is False:
            self._recons_params._set_parameter_value(parameter="NO_CHECK", value=int(b))

    def __buildZeroOffMask(self):
        self._qcbZeroRegionMask = qt.QCheckBox(
            "Set to zero the region outside the reconstruction mask", parent=self
        )
        self.linkCheckboxWithH5Variable(self._qcbZeroRegionMask, "ZEROOFFMASK")
        return self._qcbZeroRegionMask

    def _zeroRegionMaskChanged(self, b):
        if self._isLoading is False:
            self._recons_params._set_parameter_value(
                parameter="ZEROOFFMASK", value=int(b)
            )

    def __buildFixHD(self):
        self._qcbFixHD = qt.QCheckBox(
            "disable try fixed header size determination", parent=self
        )
        self.linkCheckboxWithH5Variable(self._qcbFixHD, "FIXHD")
        return self._qcbFixHD

    def _fixHDChanged(self, b):
        if self._isLoading is False:
            self._recons_params._set_parameter_value(parameter="FIXHD", value=int(b))

    def _setNumericalPart(self, val):
        self._qsbNumPart.setValue(val)

    def _getNumericalPart(self):
        return self._qsbNumPart.value()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["C. Nemoz", "H. Payno"]
__license__ = "MIT"
__date__ = "23/03/2021"

import logging
import os
from collections import OrderedDict
from silx.gui import qt
from tomwer.core.process.control.scanlist import _ScanList
from tomwer.core.utils import logconfig
from tomwer.gui.qfolderdialog import QScanDialog
from tomwer.gui.control.actions import NXTomomillParamsAction
from tomwer.gui.control.actions import CFGFileActiveLabel
from tomwer.gui.utils.inputwidget import ConfigFileSelector
from tomwer.gui.utils.inputwidget import NXTomomillOutputDirSelector
from nxtomomill import converter as nxtomomill_converter
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.blissscan import BlissScan
from tomwer.core.scan.scanfactory import ScanFactory

try:
    # new HDF5Config class
    from nxtomomill.io.config import TomoHDF5Config as HDF5Config
except ImportError:
    from nxtomomill.io.config import HDF5Config

logger = logging.getLogger(__name__)


class _DataListDialog(qt.QDialog):
    """A simple list of dataset path.BlissHDF5DataListDialog

    .. warning: the widget won't check for scan validity and will only
        emit the path to folders to the next widgets

    :param parent: the parent widget
    """

    sigUpdated = qt.Signal()
    """signal emitted when the list is updated"""

    def __init__(self, parent=None):
        qt.QDialog.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())
        # add list
        self.datalist = self.createDataList()
        self.layout().addWidget(self.datalist)
        # add buttons
        self._buttons = qt.QDialogButtonBox(parent=self)
        self._addButton = qt.QPushButton("Add", parent=self)
        self._buttons.addButton(self._addButton, qt.QDialogButtonBox.ActionRole)
        self._rmButton = qt.QPushButton("Remove", parent=self)
        self._buttons.addButton(self._rmButton, qt.QDialogButtonBox.ActionRole)
        self._rmAllButton = qt.QPushButton("Remove all", parent=self)
        self._buttons.addButton(self._rmAllButton, qt.QDialogButtonBox.ActionRole)

        self._sendButton = qt.QPushButton("Send", parent=self)
        self._buttons.addButton(self._sendButton, qt.QDialogButtonBox.AcceptRole)
        self.layout().addWidget(self._buttons)

        # expose API
        self._sendList = self.datalist._sendList
        self.setScanIDs = self.datalist.setScanIDs
        self._scheme_title = self.datalist._scheme_title
        self.length = self.datalist.length
        self.selectAll = self.datalist.selectAll
        self.clear = self.datalist.clear

        # connect signal / slot
        self._addButton.clicked.connect(self._callbackAddPath)
        self._rmButton.clicked.connect(self._callbackRemoveFolder)
        self._rmAllButton.clicked.connect(self._callbackRemoveAllFolders)

    def add(self, scan, *args, **kwargs):
        self.datalist.add(scan, *args, **kwargs)
        self.sigUpdated.emit()

    def remove(self, scan):
        self.datalist.remove(scan)
        self.sigUpdated.emit()

    def n_scan(self):
        return len(self.datalist.items)

    def _callbackAddPath(self):
        """ """
        self.sigUpdated.emit()

    def _callbackRemoveFolder(self):
        """ """
        selectedItems = self.datalist.selectedItems()
        toRemove = []
        if selectedItems is not None and len(selectedItems) > 0:
            for item in selectedItems:
                toRemove.append(item.text())
        for scan in toRemove:
            self.remove(scan)
        self.sigUpdated.emit()

    def _callbackRemoveAllFolders(self):
        self.datalist.selectAll()
        self._callbackRemoveFolder()

    def createDataList(self):
        raise NotImplementedError("Base class")


class _BlissConfigDialog(qt.QDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.setLayout(qt.QFormLayout())

        # handle configuration file
        self._configurationWidget = ConfigFileSelector()
        self.layout().addRow("configuration file", self._configurationWidget)
        # handle configuration file
        self._nxTomomillOutputWidget = NXTomomillOutputDirSelector()
        self.layout().addRow("nexus file output dir", self._nxTomomillOutputWidget)

        # buttons
        types = qt.QDialogButtonBox.Ok
        self.__buttons = qt.QDialogButtonBox(parent=self)
        self.__buttons.setStandardButtons(types)
        self.layout().addWidget(self.__buttons)

        self.__buttons.accepted.connect(self.accept)
        self.sigConfigFileChanged = self._configurationWidget.sigConfigFileChanged
        self.sigOutputdirChanged = self._nxTomomillOutputWidget.sigChanged

    def getCFGFilePath(self):
        return self._configurationWidget.getCFGFilePath()

    def setCFGFilePath(self, cfg_file):
        self._configurationWidget.setCFGFilePath(cfg_file)

    def getOutputFolder(self):
        return self._nxTomomillOutputWidget.getOutputFolder()

    def setOutputDialog(self, output_dir):
        self._nxTomomillOutputWidget.setOutputFolder(output_dir)

    def accept(self):
        self.hide()


class BlissHDF5DataListDialog(_DataListDialog):
    """Dialog used to load .h5 only (used when for nxtomomillOW when we need)
    to do a conversion from bliss.h5 to NXtomo"""

    def __init__(self, parent):
        assert isinstance(parent, BlissHDF5DataListMainWindow)
        _DataListDialog.__init__(self, parent)
        self._sendButton.setText("Send all")
        self._sendSelectedButton = qt.QPushButton("Send selected", self)
        self._buttons.addButton(
            self._sendSelectedButton, qt.QDialogButtonBox.AcceptRole
        )

    def createDataList(self):
        return BlissDataList(self)

    def _callbackAddPath(self):
        """ """
        dialog = qt.QFileDialog(self)
        dialog.setNameFilters(["HDF5 file *.h5 *.hdf5 *.nx *.nexus", "nxs"])

        if not dialog.exec_():
            dialog.close()
            return

        filesSelected = dialog.selectedFiles()
        for file_ in filesSelected:
            self.add(file_, configuration=self.parent().getHDF5Config())
        super()._callbackAddPath()


class BlissHDF5DataListMainWindow(qt.QMainWindow):

    sigNXTomoCFGFileChanged = qt.Signal(str)
    """signal emitted when the configuration file change"""

    sigUpdated = qt.Signal()
    """signal emitted when the list of bliss scan to convert change"""

    def __init__(self, parent):
        super().__init__(parent)
        self._widget = BlissHDF5DataListDialog(self)
        # rework BlissHDF5DataListDialog layout
        self._subWidget = qt.QWidget(self)
        self._subWidget.setLayout(qt.QVBoxLayout())
        self._subWidget.layout().addWidget(self._widget.datalist)
        self._subWidget.layout().addWidget(self._widget._buttons)
        self.setCentralWidget(self._subWidget)

        self._dialog = _BlissConfigDialog(self)
        self._dialog.setWindowTitle("Select nxtomomill configuration file")

        # add toolbar
        toolbar = qt.QToolBar(self)
        toolbar.setMovable(False)
        toolbar.setFloatable(False)
        self.addToolBar(qt.Qt.TopToolBarArea, toolbar)

        # add filtering
        self._parametersAction = NXTomomillParamsAction(toolbar)
        toolbar.addAction(self._parametersAction)
        self._parametersAction.triggered.connect(self._parametersTriggered)

        # toolbar spacer
        spacer = qt.QWidget(toolbar)
        spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        toolbar.addWidget(spacer)

        # add information if cfg file is activate or not
        self._cfgStatusLabel = CFGFileActiveLabel(toolbar)
        toolbar.addWidget(self._cfgStatusLabel)

        # expose API
        self.getCFGFilePath = self._dialog.getCFGFilePath
        self.setCFGFilePath = self._dialog.setCFGFilePath
        self.setScanIDs = self._widget.setScanIDs
        self._sendButton = self._widget._sendButton
        self._sendSelectedButton = self._widget._sendSelectedButton
        self.add = self._widget.add
        self.n_scan = self._widget.n_scan
        self.datalist = self._widget.datalist

        # connect signal / slot
        self._dialog.sigConfigFileChanged.connect(self._cfgFileChanged)
        self._dialog.sigOutputdirChanged.connect(self._propagateUpdateSignal)
        self._widget.sigUpdated.connect(self._propagateUpdateSignal)

    def getOutputFolder(self):
        return self._dialog.getOutputFolder()

    def setOutputFolder(self, output_dir):
        self._dialog.setOutputDialog(output_dir)

    def _propagateUpdateSignal(self):
        self.sigUpdated.emit()

    def _parametersTriggered(self):
        self._dialog.show()
        self._dialog.raise_()

    def _cfgFileChanged(self):
        cfg_file = self._dialog.getCFGFilePath()
        if cfg_file in (None, "") or not os.path.exists(cfg_file):
            self._cfgStatusLabel.setInactive()
        else:
            try:
                HDF5Config.from_cfg_file(cfg_file)
            except Exception:
                self._cfgStatusLabel.setInactive()
            else:
                self._cfgStatusLabel.setActive()
        if cfg_file is None:
            cfg_file = ""
        self.sigNXTomoCFGFileChanged.emit(cfg_file)

    def getHDF5Config(self):
        """Return default HDF5Config or the one created from use input"""
        cfg_file = self._dialog.getCFGFilePath()
        if cfg_file in (None, ""):
            return HDF5Config()
        else:
            try:
                config = HDF5Config.from_cfg_file(cfg_file)
            except Exception:
                return HDF5Config()
            else:
                return config


class GenericDataListDialog(_DataListDialog):
    """Dialog used to load EDFScan or HEDF5 scans"""

    def createDataList(self):
        return GenericDataList(self)

    def _callbackAddPath(self):
        """ """
        dialog = QScanDialog(self, multiSelection=True)
        dialog.setNameFilters(
            [
                "HDF5 files (*.h5 *.hdf5 *.nx *.nexus)",
                "Nexus files (*.nx *.nexus)",
                "Any files (*)",
            ]
        )

        if not dialog.exec_():
            dialog.close()
            return

        files_or_folders = dialog.files_selected()
        for file_or_folder in files_or_folders:
            self.add(file_or_folder)
        super()._callbackAddPath()


class _DataList(_ScanList, qt.QTableWidget):
    def __init__(self, parent):
        _ScanList.__init__(self)
        self._target = None
        qt.QTableWidget.__init__(self, parent)
        self.setRowCount(0)
        self.setColumnCount(1)
        self.setSortingEnabled(True)
        self.verticalHeader().hide()
        if hasattr(self.horizontalHeader(), "setSectionResizeMode"):  # Qt5
            self.horizontalHeader().setSectionResizeMode(0, qt.QHeaderView.Stretch)
        else:  # Qt4
            self.horizontalHeader().setResizeMode(0, qt.QHeaderView.Stretch)
        self.setAcceptDrops(True)
        self.items = OrderedDict()

        # QMenu
        self.menu = qt.QMenu(self)
        self._copyAction = qt.QAction("copy")
        self._copyAction.triggered.connect(self._copyRequested)
        self.menu.addAction(self._copyAction)

    def remove_item(self, item):
        """Remove a given folder"""
        del self.items[item.text()]
        itemIndex = self.row(item)
        self.takeItem(itemIndex, 0)
        _ScanList.remove(self, item.text())
        self.removeRow(item.row())
        self.setRowCount(self.rowCount() - 1)
        self._update()

    def remove(self, scan):
        if scan is not None and scan in self.items:
            item = self.items[scan]
            itemIndex = self.row(item)
            self.takeItem(itemIndex, 0)
            _ScanList.remove(self, scan)
            del self.items[scan]
            self.removeRow(item.row())
            self.setRowCount(self.rowCount() - 1)
            self._update()

    def _update(self):
        list_scan = list(self.items.keys())
        self.clear()
        for scan in list_scan:
            self.add(scan)
        self.sortByColumn(0, self.horizontalHeader().sortIndicatorOrder())

    def add(self, d):
        """add the folder or path"""
        raise NotImplementedError("Base class")

    def _addScanIDItem(self, d):
        if "@" not in d and not os.path.isdir(d):
            warning = (
                "Skipping the observation of %s, directory not existing on the system"
                % d
            )
            logger.info(warning, extra={logconfig.DOC_TITLE: self._scheme_title})
        elif d in self.items:
            logger.debug("The path {} is already in the scan list".format(d))
        else:
            row = self.rowCount()
            self.setRowCount(row + 1)

            _item = qt.QTableWidgetItem()
            _item.setText(d)
            _item.setFlags(qt.Qt.ItemIsEnabled | qt.Qt.ItemIsSelectable)
            self.setItem(row, 0, _item)

            self.items[d] = _item

    def setScanIDs(self, scanIDs):
        [self._addScanIDItem(item) for item in scanIDs]
        _ScanList.setScanIDs(self, scanIDs)

    def clear(self):
        """Remove all items on the list"""
        self.items = OrderedDict()
        _ScanList.clear(self)
        qt.QTableWidget.clear(self)
        self.setRowCount(0)
        self.setHorizontalHeaderLabels(["folder"])
        if hasattr(self.horizontalHeader(), "setSectionResizeMode"):  # Qt5
            self.horizontalHeader().setSectionResizeMode(0, qt.QHeaderView.Stretch)
        else:  # Qt4
            self.horizontalHeader().setResizeMode(0, qt.QHeaderView.Stretch)

    def dropEvent(self, event):
        if event.mimeData().hasFormat("text/uri-list"):
            for url in event.mimeData().urls():
                self.add(str(url.path()))

    def supportedDropActions(self):
        """Inherited method to redefine supported drop actions."""
        return qt.Qt.CopyAction | qt.Qt.MoveAction

    def dragEnterEvent(self, event):
        if event.mimeData().hasFormat("text/uri-list"):
            event.accept()
            event.setDropAction(qt.Qt.CopyAction)
        else:
            qt.QListWidget.dragEnterEvent(self, event)

    def dragMoveEvent(self, event):
        if event.mimeData().hasFormat("text/uri-list"):
            event.setDropAction(qt.Qt.CopyAction)
            event.accept()
        else:
            qt.QListWidget.dragMoveEvent(self, event)

    def _scanAt(self, point):
        item = self.itemAt(point)
        if item is not None:
            return item.text()

    def contextMenuEvent(self, event):
        self._target = self._scanAt(event.pos())
        if self._target is not None:
            self.menu.exec_(event.globalPos())

    def _copyRequested(self):
        clipboard = qt.QGuiApplication.clipboard()
        clipboard.setText(self._target)


class GenericDataList(_DataList):
    """Data list able to manage directories (EDF/HDF5?) or files (HDF5)"""

    def __init__(self, parent):
        _DataList.__init__(self, parent)
        self.setHorizontalHeaderLabels(["scan"])

    def add(self, path):
        """Add the path folder d in the scan list

        :param d: the path of the directory to add
        :type d: Union[str, TomoBase]
        """
        if isinstance(path, TomwerScanBase):
            scan_obj = _ScanList.add(self, path)
            self._addScanIDItem(str(scan_obj))
        elif os.path.isdir(path):
            scan_obj = _ScanList.add(self, path)
            if scan_obj:
                self._addScanIDItem(str(scan_obj.path))
        else:
            # try to add it as HDF5TomoScan
            try:
                scan_objs = ScanFactory.create_scan_objects(path)
            except Exception as e:
                logger.error(e)
            else:
                if scan_objs is None:
                    return
                for scan in scan_objs:
                    scan_obj = _ScanList.add(self, scan)
                    if scan_obj:
                        self._addScanIDItem(str(scan_obj))


class BlissDataList(_DataList):
    def __init__(self, parent):
        _DataList.__init__(self, parent)
        self.setHorizontalHeaderLabels(["entry@bliss_file.h5"])
        self._configurations = {}

    def _update(self):
        list_scan = list(self.items.keys())
        self.clear()
        for scan in list_scan:
            self.add(scan, configuration=self._configurations.get(scan, None))
        self.sortByColumn(0, self.horizontalHeader().sortIndicatorOrder())

    def remove(self, scan):
        if scan is not None and scan in self.items:
            if str(scan) in self._configurations:
                del self._configurations[str(scan)]
        super().remove(scan=scan)

    def add(self, path, configuration):
        """Add the path folder d in the scan list

        :param d: the path of the directory to add
        :type d: Union[str, TomoBase]
        """
        if "@" in path:
            entry, path = path.split("@")
            possible_entries = [entry]
        else:
            if not BlissScan.is_bliss_file(path):
                msg = qt.QMessageBox(self)
                msg.setIcon(qt.QMessageBox.Warning)
                types = qt.QMessageBox.Ok | qt.QMessageBox.Cancel
                msg.setStandardButtons(types)

                if HDF5TomoScan.is_nexus_nxtomo_file(path):
                    text = (
                        "The input file `{}` seems to contain `NXTomo` entries. "
                        "and no valid `Bliss` valid entry. \n"
                        "This is probably not a Bliss file. Do you still want to "
                        "translate ?".format(path)
                    )
                else:
                    text = (
                        "The input file `{}` does not seems to contain any "
                        "valid `Bliss` entry. \n"
                        "This is probably not a Bliss file. Do you still want to "
                        "translate ?".format(path)
                    )
                msg.setText(text)
                if msg.exec_() != qt.QMessageBox.Ok:
                    return

            try:
                possible_entries = nxtomomill_converter.get_bliss_tomo_entries(
                    path, configuration
                )
            except Exception:
                logger.error("Faild to find entries for {}".format(path))
                possible_entries = []

        for entry in possible_entries:
            scan = HDF5TomoScan(scan=path, entry=entry)
            scan_obj = _ScanList.add(self, scan)
            if scan_obj:
                self._configurations[str(scan_obj)] = configuration
                self._addScanIDItem(str(scan_obj))



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/


__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "27/02/2019"


import h5py
from tomoscan.scanbase import TomoScanBase
from tomoscan.scanfactory import ScanFactory as _oScanFactory
from tomoscan.esrf.hdf5scan import HDF5TomoScan as _oHDF5TomoScan
from tomwer.utils import docstring
from .scanbase import TomwerScanBase
from .edfscan import EDFTomoScan
from .hdf5scan import HDF5TomoScan
from .blissscan import BlissScan
import os
import glob
import json
import logging

_logger = logging.getLogger(__name__)


@docstring(_oScanFactory)
class ScanFactory(object):
    @docstring(_oScanFactory.create_scan_object)
    @staticmethod
    def create_scan_object(scan_path, entry=None, accept_bliss_scan=False):
        """

        :param TextIOWrapper scan_path: path to the scan directory or file
        :param entry: entry on the file. Requested for hdf5 files
        :param accept_bliss_scan: if True the factory can return some BlissScan
                                  But this is only compatible with the
                                  Tomomill processing.
        :return: TomwerScanBase instance fitting the scan folder or scan path
        :rtype: tomwer.core.scan.scanbase.TomwerScanBase
        :raises: ValueError if scan_path is not containing a scan
        """
        if scan_path is None:
            raise ValueError("'scan_path' should be provided")
        if ScanFactory.is_hdf5_tomo(scan_path):
            valid_entries = _oHDF5TomoScan.get_valid_entries(scan_path)
            if entry is None:
                if len(valid_entries) > 1:
                    _logger.warning(
                        "more than one entry found for %s."
                        "Pick the last entry" % scan_path
                    )
                entry = valid_entries[-1]
            elif entry not in valid_entries:
                raise ValueError(
                    "entry {} is invalid. Does it exists ? Is the "
                    "file NXTomo compliant ?. Valid entry are {}"
                    "".format(entry, valid_entries)
                )
            return HDF5TomoScan(scan=scan_path, entry=entry)
        elif ScanFactory.is_edf_tomo(scan_path):
            return EDFTomoScan(scan=scan_path)
        elif accept_bliss_scan and BlissScan.is_bliss_file(scan_path):
            if entry is None:
                valid_entries = BlissScan.get_valid_entries(scan_path)
                if len(valid_entries) > 1:
                    _logger.warning(
                        "more than one entry found for %s."
                        "Pick the last entry" % scan_path
                    )
                entry = valid_entries[-1]
            return BlissScan(master_file=scan_path, entry=entry, proposal_file=None)

        raise ValueError("Unable to generate a scan object from %s" % scan_path)

    @docstring(_oScanFactory.create_scan_objects)
    @staticmethod
    def create_scan_objects(scan_path, accept_bliss_scan=True) -> tuple:
        scan_path = scan_path.rstrip(os.path.sep)
        if EDFTomoScan.is_tomoscan_dir(scan_path):
            return (EDFTomoScan(scan=scan_path),)
        elif HDF5TomoScan.is_tomoscan_dir(scan_path):
            scans = []
            master_file = HDF5TomoScan.get_master_file(scan_path=scan_path)
            entries = HDF5TomoScan.get_valid_entries(master_file)
            for entry in entries:
                scans.append(HDF5TomoScan(scan=scan_path, entry=entry, index=None))
            return tuple(scans)
        elif accept_bliss_scan and BlissScan.is_bliss_file(scan_path):
            scans = []
            for entry in BlissScan.get_valid_entries(scan_path):
                scans.append(
                    BlissScan(master_file=scan_path, entry=entry, proposal_file=None)
                )
            return tuple(scans)
        return tuple()

    @staticmethod
    def mock_scan(type="edf"):
        """Mock a scan structure which is not associated to any real acquistion"""
        if type == "edf":
            return EDFTomoScan(scan=None)
        else:
            raise NotImplementedError("Other TomoScan are not defined yet")

    @staticmethod
    def create_scan_object_frm_dict(_dict):
        if TomoScanBase.DICT_TYPE_KEY not in _dict:
            raise ValueError(
                "given dict is not recognized. Cannot find {}".format(
                    TomoScanBase.DICT_TYPE_KEY
                )
            )
        elif _dict[TomoScanBase.DICT_TYPE_KEY] == EDFTomoScan._TYPE:
            return EDFTomoScan(scan=None).load_from_dict(_dict)
        elif _dict[TomoScanBase.DICT_TYPE_KEY] == HDF5TomoScan._TYPE:
            return HDF5TomoScan.from_dict(_dict)
        else:
            raise ValueError(
                "Scan type {} is not managed".format(_dict[TomoScanBase.DICT_TYPE_KEY])
            )

    @staticmethod
    def is_tomo_scandir(scan_path):
        """

        :param str scan_path: path to the scan directory or file
        :return: True if the given path / file is a tomo_scandir. For now yes by
                 default
        :rtype: bool
        """
        return True

    @staticmethod
    def is_edf_tomo(scan_path):
        """

        :param str scan_path: path to the scan directory or file
        :return: True if given path define a tomo scan based on .edf file
        :rtype: bool
        """
        if scan_path and os.path.isdir(scan_path):
            file_basename = os.path.basename(scan_path)
            has_info_file = (
                len(glob.glob(os.path.join(scan_path, file_basename + "*.info"))) > 0
            )
            not_lbs_scan_path = scan_path.replace("lbsram", "", 1)
            has_notlbsram_info_file = (
                len(
                    glob.glob(os.path.join(not_lbs_scan_path, file_basename + "*.info"))
                )
                > 0
            )
            if has_info_file or has_notlbsram_info_file:
                return True
        return False

    @staticmethod
    def is_hdf5_tomo(scan_path):
        """

        :param scan_path:
        :return:
        """
        if os.path.isfile(scan_path):
            return h5py.is_hdf5(scan_path)
        else:
            return HDF5TomoScan.directory_contains_scan(scan_path)

    @staticmethod
    def create_from_json(desc):
        """Create a ScanBase instance from a json description"""
        data = json.load(desc)

        if TomwerScanBase._DICT_TYPE_KEY not in data:
            raise ValueError("json not recognize")
        elif data[TomwerScanBase._DICT_TYPE_KEY] == EDFTomoScan._TYPE:
            scan = EDFTomoScan(scan=None).load_from_dict(data)
            return scan
        else:
            raise ValueError("Type", data[TomwerScanBase.type], "is not managed")



# # coding: utf-8
# # /*##########################################################################
# #
# # Copyright (c) 2016-2017 European Synchrotron Radiation Facility
# #
# # Permission is hereby granted, free of charge, to any person obtaining a copy
# # of this software and associated documentation files (the "Software"), to deal
# # in the Software without restriction, including without limitation the rights
# # to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# # copies of the Software, and to permit persons to whom the Software is
# # furnished to do so, subject to the following conditions:
# #
# # The above copyright notice and this permission notice shall be included in
# # all copies or substantial portions of the Software.
# #
# # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# # IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# # FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# # AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# # LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# # THE SOFTWARE.
# #
# # ###########################################################################*/
# """ Not a unit test but send a simple test message error to gray log"""
#
# __authors__ = ["H. Payno"]
# __license__ = "MIT"
# __date__ = "24/01/2017"
#
# from tomwer.web.client import OWClient
# import logging
#
# logger = logging.getLogger(__name__)
#
#
# class TestGrayLog(OWClient):
#     def __init__(self):
#         OWClient.__init__(self)
#
#     def sendErrorMessage(self):
#         logger.error("test error message")
#
#     def sendwarningMessage(self):
#         logger.warning("test warning message")
#
#     def sendInfoMessage(self):
#         logger.info("test info message")
#
#     def sendProcessEndedMessage(self):
#         logger.processEnded("test processEnded message")
#
#
# if __name__ == "__main__":
#     c = TestGrayLog()
#     c.sendErrorMessage()
#     c.sendwarningMessage()
#     c.sendInfoMessage()
#     c.sendProcessEndedMessage()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "15/12/2021"

from typing import Optional

from processview.core.manager.manager import DatasetState, ProcessManager
from tomwer.core import settings
from orangecontrib.tomwer.widgets import utils
from ...orange.managedprocess import SuperviseOW
from orangecontrib.tomwer.widgets.utils import WidgetLongProcessing
from orangewidget.settings import Setting
from orangewidget import gui
from orangewidget.widget import Input, Output
from silx.gui import qt
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.synctools.stacks.reconstruction.castvolume import CastVolumeProcessStack
import tomwer.core.process.reconstruction.nabu.castvolume
from tomwer.gui.reconstruction.nabu.castvolume import CastVolumeWidget
import logging

_logger = logging.getLogger(__name__)


class CastNabuVolumeOW(WidgetLongProcessing, SuperviseOW):
    """
    widget used to cast from 32 bits tiff to 16 bits tiff.

    This is done in a separate process because:

    * this is done in cpu when nabu reconstruct on GPU and this should free sooner GPU ressources with the current architecture.
    * this is not included in nabu but also done as post processing.
    * limitation is that not having computed the histogram during volume construction will slow down the cast

    :param parent: the parent widget
    """

    # note of this widget should be the one registered on the documentation
    name = "cast volume"
    id = "orange.widgets.tomwer.reconstruction.CastNabuVolumeOW.CastNabuVolumeOW"
    description = "This widget will allow to cast from 32 bits to 16 bits nabu volume. Only works for tiff now"
    icon = "icons/nabu_cast.svg"
    priority = 60
    keywords = [
        "tomography",
        "nabu",
        "reconstruction",
        "volume",
        "cast",
        "tiff",
        "32 bits",
        "16 bits",
        "tif",
    ]

    ewokstaskclass = tomwer.core.process.reconstruction.nabu.castvolume.CastVolumeTask

    want_main_area = True
    resizing_enabled = True
    compress_signal = False
    allows_cycle = True

    _rpSetting = Setting(dict())
    # kept for compatibility

    static_input = Setting({"data": None, "cast_volume_params": {}})

    sigScanReady = qt.Signal(TomwerScanBase)
    "Signal emitted when a scan is ended"

    TIMEOUT = 30

    class Inputs:
        data = Input(
            name="data",
            type=TomwerScanBase,
            doc="one scan to be process",
            default=True,
            multiple=False,
        )

    class Outputs:
        data = Output(name="data", type=TomwerScanBase, doc="one scan to be process")

    def __init__(self, parent=None):
        """ """
        SuperviseOW.__init__(self, parent)
        WidgetLongProcessing.__init__(self)
        # processing tool
        self._processingStack = CastVolumeProcessStack(process_id=self.process_id)
        self._window = CastVolumeWidget(parent=self)

        self._layout = gui.vBox(self.mainArea, self.name).layout()
        self._layout.setContentsMargins(0, 0, 0, 0)

        self._layout.addWidget(self._window)

        self.setConfiguration(self.static_input.get("cast_volume_params", {}))

        # connect signal / slot
        self._window.sigConfigChanged.connect(self._updateConfig)
        # connect signal / slot
        self._processingStack.sigComputationStarted.connect(self._startProcessing)
        self._processingStack.sigComputationEnded.connect(self._endProcessing)

    @Inputs.data
    def process(self, scan: Optional[TomwerScanBase]):
        if scan is None:
            return
        if not isinstance(scan, TomwerScanBase):
            raise TypeError(
                f"scan is expected to be an instance of {TomwerScanBase} not {type(scan)}"
            )
        if (
            settings.isOnLbsram(scan)
            and utils.isLowOnMemory(settings.get_lbsram_path()) is True
        ):
            ProcessManager().notify_dataset_state(
                dataset=scan,
                process=self,
                state=DatasetState.SKIPPED,
            )
            _logger.processSkipped(
                f"skip volume cast for {str(scan)} because lbsram is almost full. Try to free it"
            )
            self.Outputs.data.send(scan)
        else:
            self._processingStack.add(scan, configuration=self.getConfiguration())

    def getConfiguration(self) -> dict:
        return self._window.getConfiguration()

    def setConfiguration(self, configuration: dict) -> None:
        self._window.setConfiguration(configuration)

    def _updateConfig(self):
        self.static_input = {
            "data": None,
            "cast_volume_params": self.getConfiguration(),
        }

    def _endProcessing(self, scan, future_scan):
        WidgetLongProcessing._endProcessing(self, scan)
        if scan is not None:
            self.Outputs.data.send(scan)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "23/10/2019"


from .base import _ReconsParam
import logging

_logger = logging.getLogger(__name__)


class AxisRP(_ReconsParam):
    """
    Define axis parameters for ftserie reconstruction
    """

    AXIS_POSITION_PAR_KEY = "ROTATION_AXIS_POSITION"
    """Key used for the axis position in par files"""

    def __init__(self):
        super().__init__()
        self.__do_axis_correction = True
        self.__use_tomwer_axis = True
        self.__use_old_tomwer_axis = True

        # binding
        self._managed_params = {
            "DO_AXIS_CORRECTION": self.__class__.do_axis_correction,
            "USE_TOMWER_AXIS": self.__class__.use_tomwer_axis,
            "TRY_USE_OLD_TOMWER_AXIS": self.__class__.use_old_tomwer_axis,
        }

    @property
    def do_axis_correction(self):
        """Activate pyhst axis translation correction"""
        return self.__do_axis_correction

    @do_axis_correction.setter
    def do_axis_correction(self, do_axis_correction):
        do_axis_correction = bool(do_axis_correction)
        if do_axis_correction != self.__do_axis_correction:
            self.__do_axis_correction = do_axis_correction
            self.changed()

    @property
    def use_tomwer_axis(self):
        """If the scan contains a valid axis_params.value then use it."""
        return self.__use_tomwer_axis

    @use_tomwer_axis.setter
    def use_tomwer_axis(self, use_tomwer_axis):
        use_tomwer_axis = bool(use_tomwer_axis)
        if self.__use_tomwer_axis != use_tomwer_axis:
            self.__use_tomwer_axis = use_tomwer_axis
            self.changed()

    @property
    def use_old_tomwer_axis(self):
        """If the scan does not contains an axis_params but contains a
        [XXX]tomwer_processes.h5 file from a previous computation, try to
        get the center of rotation from it"""
        return self.__use_old_tomwer_axis

    @use_old_tomwer_axis.setter
    def use_old_tomwer_axis(self, use_old_tomwer_axis):
        use_old_tomwer_axis = bool(use_old_tomwer_axis)
        if self.__use_old_tomwer_axis != use_old_tomwer_axis:
            self.__use_old_tomwer_axis = bool(use_old_tomwer_axis)
            self.changed()

    def to_dict(self):
        return {
            "DO_AXIS_CORRECTION": str(int(self.do_axis_correction)),
            "USE_TOMWER_AXIS": str(int(self.use_tomwer_axis)),
            "TRY_USE_OLD_TOMWER_AXIS": str(int(self.use_tomwer_axis)),
        }

    @staticmethod
    def from_dict(_dict):
        axis = AxisRP()
        axis.load_from_dict(_dict=_dict)
        return axis

    def load_from_dict(self, _dict):
        try:
            self.do_axis_correction = int(_dict["DO_AXIS_CORRECTION"])
        except KeyError:
            _logger.info('no "DO_AXIS_CORRECTION" key')
        try:
            self.use_tomwer_axis = int(_dict["USE_TOMWER_AXIS"])
        except KeyError:
            _logger.info('no "USE_TOMWER_AXIS" key')
        try:
            self.use_old_tomwer_axis = int(_dict["TRY_USE_OLD_TOMWER_AXIS"])
        except KeyError:
            _logger.info('no "TRY_USE_OLD_TOMWER_AXIS" key')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import sys
from silx.gui import qt
import argparse
from tomwer.gui import icons
from tomwer.gui.utils.splashscreen import getMainSplashScreen
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomoscan.esrf.hdf5scan import ImageKey
from tomwer.gui.edit.imagekeyeditor import ImageKeyDialog as _ImageKeyDialog
from nxtomomill.utils import change_image_key_control
from tomwer.core.utils.resource import increase_max_number_file
import signal

logging.basicConfig(level=logging.WARNING)
_logger = logging.getLogger(__name__)


class ImageKeyDialog(_ImageKeyDialog):
    def __init__(self, *args, **kwargs):
        _ImageKeyDialog.__init__(self, *args, **kwargs)
        self._scan = None

    def setScan(self, scan):
        if not isinstance(scan, HDF5TomoScan):
            raise TypeError("This only manage HDF5TomoScan")
        self._scan = scan
        _ImageKeyDialog.setScan(self, scan)

    def _validateReleased(self):
        if not self._scan:
            raise ValueError("Scan should be set first")

        modifications = self.getModifications()
        image_keys_set = set(self.getModifications().values())
        image_keys_set = set(
            [ImageKey.from_value(image_key) for image_key in image_keys_set]
        )
        for image_key_type in image_keys_set:
            frame_indexes_dict = dict(
                filter(lambda item: item[1] is image_key_type, modifications.items())
            )
            frame_indexes = tuple(frame_indexes_dict.keys())
            _logger.warning(
                "will modify {} to {}".format(frame_indexes, image_key_type)
            )
            change_image_key_control(
                file_path=self._scan.master_file,
                entry=self._scan.entry,
                frames_indexes=frame_indexes,
                image_key_control_value=image_key_type.value,
                logger=_logger,
            )


def getinputinfo():
    return "tomwer nabu [scan_path]"


def sigintHandler(*args):
    """Handler for the SIGINT signal."""
    qt.QApplication.quit()


def main(argv):
    import os

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("scan_path", help="HDF5File to edit")
    parser.add_argument("entry", default=None, help="Entry to treat")
    options = parser.parse_args(argv[1:])

    # image key can only handle HDF5Tomoscan for now
    scan = HDF5TomoScan(scan=options.scan_path, entry=options.entry)
    increase_max_number_file()

    app = qt.QApplication.instance() or qt.QApplication([])

    qt.QLocale.setDefault(qt.QLocale(qt.QLocale.English))
    qt.QLocale.setDefault(qt.QLocale.c())
    signal.signal(signal.SIGINT, sigintHandler)
    sys.excepthook = qt.exceptionHandler
    timer = qt.QTimer()
    timer.start(500)
    # Application have to wake up Python interpreter, else SIGINT is not
    # catch
    timer.timeout.connect(lambda: None)

    splash = getMainSplashScreen()
    options.scan_path = options.scan_path.rstrip(os.path.sep)

    widget = ImageKeyDialog(parent=None)
    widget.setScan(scan)
    widget.setWindowTitle("Image key editor")
    widget.setWindowIcon(icons.getQIcon("tomwer"))
    splash.finish(widget)
    widget.show()
    app.exec_()


if __name__ == "__main__":
    main(sys.argv)
